# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot import config
from buildbot.interfaces import IRenderable
from buildbot.process.buildstep import BuildStep
from buildbot.process.buildstep import ShellMixin


class CMake(ShellMixin, BuildStep):
    DEFAULT_CMAKE = 'cmake'

    name = 'cmake'
    description = ['running', 'cmake']
    descriptionDone = ['cmake']

    renderables = ('cmake', 'definitions', 'generator', 'options', 'path')

    haltOnFailure = True

    def __init__(
        self,
        path=None,
        generator=None,
        definitions=None,
        options=None,
        cmake=DEFAULT_CMAKE,
        **kwargs,
    ):
        self.path = path
        self.generator = generator

        if not (
            definitions is None
            or isinstance(definitions, dict)
            or IRenderable.providedBy(definitions)
        ):
            config.error('definitions must be a dictionary or implement IRenderable')
        self.definitions = definitions

        if not (
            options is None or isinstance(options, (list, tuple)) or IRenderable.providedBy(options)
        ):
            config.error('options must be a list, a tuple or implement IRenderable')
        self.options = options

        self.cmake = cmake
        kwargs = self.setupShellMixin(kwargs, prohibitArgs=['command'])
        super().__init__(**kwargs)

    @defer.inlineCallbacks
    def run(self):
        """
        run CMake
        """
        command = [self.cmake]

        if self.generator:
            command.extend(['-G', self.generator])

        if self.definitions is not None:
            for item in self.definitions.items():
                command.append(f'-D{item[0]}={item[1]}')

        if self.options is not None:
            command.extend(self.options)

        if self.path:
            command.append(self.path)

        cmd = yield self.makeRemoteShellCommand(command=command)

        yield self.runCommand(cmd)

        return cmd.results()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from unittest import mock

from twisted.internet import defer


class PBManagerMixin:
    def setUpPBChangeSource(self):
        "Set up a fake self.pbmanager."
        self.registrations = []
        self.unregistrations = []
        pbm = self.pbmanager = mock.Mock()
        pbm.register = self._fake_register

    def _fake_register(self, portstr, username, password, factory):
        reg = mock.Mock()

        def unregister():
            self.unregistrations.append((portstr, username, password))
            return defer.succeed(None)

        reg.unregister = unregister
        self.registrations.append((portstr, username, password))
        return reg

    def assertNotRegistered(self):
        self.assertEqual(self.registrations, [])

    def assertNotUnregistered(self):
        self.assertEqual(self.unregistrations, [])

    def assertRegistered(self, portstr, username, password):
        for ps, un, pw in self.registrations:
            if ps == portstr and username == un and pw == password:
                return
        self.fail(f"not registered: {(portstr, username, password)!r} not in {self.registrations}")

    def assertUnregistered(self, portstr, username, password):
        for ps, un, pw in self.unregistrations:
            if ps == portstr and username == un and pw == password:
                return
        self.fail("still registered")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from __future__ import annotations

import json
import os
import typing
from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer
from twisted.web.error import Error

from buildbot.interfaces import IConfigured
from buildbot.util import unicode2bytes
from buildbot.www import resource

if TYPE_CHECKING:
    from buildbot.master import BuildMaster


def get_environment_versions() -> list[tuple[str, str]]:
    import sys  # pylint: disable=import-outside-toplevel

    import twisted  # pylint: disable=import-outside-toplevel

    from buildbot import version as bbversion  # pylint: disable=import-outside-toplevel

    pyversion = '.'.join(map(str, sys.version_info[:3]))

    tx_version_info = (twisted.version.major, twisted.version.minor, twisted.version.micro)
    txversion = '.'.join(map(str, tx_version_info))

    return [
        ('Python', pyversion),
        ('Buildbot', bbversion),
        ('Twisted', txversion),
    ]


def get_www_frontend_config_dict(master: BuildMaster, www_config: dict[str, Any]) -> dict[str, Any]:
    # This config is shared with the frontend.
    config = dict(www_config)

    versions = get_environment_versions()
    vs = config.get('versions')
    if isinstance(vs, list):
        versions += vs
    config['versions'] = versions

    config['buildbotURL'] = master.config.buildbotURL
    config['title'] = master.config.title
    config['titleURL'] = master.config.titleURL
    config['multiMaster'] = master.config.multiMaster

    # delete things that may contain secrets
    config.pop('change_hook_dialects', None)

    # delete things that may contain information about the serving host
    config.pop('custom_templates_dir', None)

    return config


def serialize_www_frontend_config_dict_to_json(config: dict[str, Any]) -> str:
    def to_json(obj: Any) -> dict[str, Any] | str:
        obj = IConfigured(obj).getConfigDict()
        if isinstance(obj, dict):
            return obj
        # don't leak object memory address
        obj = obj.__class__.__module__ + "." + obj.__class__.__name__
        return repr(obj) + " not yet IConfigured"

    return json.dumps(config, default=to_json)


_known_theme_variables = (
    ("bb-sidebar-background-color", "#30426a"),
    ("bb-sidebar-header-background-color", "#273759"),
    ("bb-sidebar-header-text-color", "#fff"),
    ("bb-sidebar-title-text-color", "#627cb7"),
    ("bb-sidebar-footer-background-color", "#273759"),
    ("bb-sidebar-button-text-color", "#b2bfdc"),
    ("bb-sidebar-button-hover-background-color", "#1b263d"),
    ("bb-sidebar-button-hover-text-color", "#fff"),
    ("bb-sidebar-button-current-background-color", "#273759"),
    ("bb-sidebar-button-current-text-color", "#b2bfdc"),
    ("bb-sidebar-stripe-hover-color", "#e99d1a"),
    ("bb-sidebar-stripe-current-color", "#8c5e10"),
)


def serialize_www_frontend_theme_to_css(config: dict[str, Any], indent: int) -> str:
    theme_config = config.get('theme', {})

    return ('\n' + ' ' * indent).join([
        f'--{name}: {theme_config.get(name, default)};' for name, default in _known_theme_variables
    ])


def replace_placeholder_range(string: str, start: str, end: str, replacement: str) -> str:
    # Simple string replacement is much faster than a multiline regex
    i1 = string.find(start)
    i2 = string.find(end)
    if i1 < 0 or i2 < 0:
        return string
    return string[0:i1] + replacement + string[i2 + len(end) :]


class ConfigResource(resource.Resource):
    needsReconfig = True

    def reconfigResource(self, new_config: Any) -> None:
        self.frontend_config = get_www_frontend_config_dict(self.master, new_config.www)

    def render_GET(self, request: Any) -> int:
        return self.asyncRenderHelper(request, self.do_render)

    def do_render(self, request: Any) -> defer.Deferred:
        config: dict[str, Any] = {}
        request.setHeader(b"content-type", b'application/json')
        request.setHeader(b"Cache-Control", b"public,max-age=0")

        config.update(self.frontend_config)
        config.update({"user": self.master.www.getUserInfos(request)})

        return defer.succeed(
            unicode2bytes(serialize_www_frontend_config_dict_to_json(config), encoding='ascii')
        )


class IndexResource(resource.Resource):
    # enable reconfigResource calls
    needsReconfig = True

    def __init__(self, master: BuildMaster, staticdir: str) -> None:
        super().__init__(master)
        self.static_dir = staticdir
        with open(os.path.join(self.static_dir, 'index.html')) as index_f:
            self.index_template = index_f.read()

    def reconfigResource(self, new_config: Any) -> None:
        self.config = new_config.www
        self.frontend_config = get_www_frontend_config_dict(self.master, self.config)

    def render_GET(self, request: Any) -> int:
        return self.asyncRenderHelper(request, self.renderIndex)

    @defer.inlineCallbacks
    def renderIndex(self, request: Any) -> typing.Generator[Any, None, bytes]:
        config = {}
        request.setHeader(b"content-type", b'text/html')
        request.setHeader(b"Cache-Control", b"public,max-age=0")

        try:
            yield self.config['auth'].maybeAutoLogin(request)
        except Error as e:
            config["on_load_warning"] = e.message

        config.update(self.frontend_config)
        config.update({"user": self.master.www.getUserInfos(request)})

        serialized_config = serialize_www_frontend_config_dict_to_json(config)
        serialized_css = serialize_www_frontend_theme_to_css(config, indent=8)
        rendered_index = self.index_template.replace(
            ' <!-- BUILDBOT_CONFIG_PLACEHOLDER -->',
            f"""<script id="bb-config">
    window.buildbotFrontendConfig = {serialized_config};
</script>""",
        )

        rendered_index = replace_placeholder_range(
            rendered_index,
            '<!-- BUILDBOT_THEME_CSS_PLACEHOLDER_BEGIN -->',
            '<!-- BUILDBOT_THEME_CSS_PLACEHOLDER_END -->',
            f"""<style>
      :root {{
        {serialized_css}
      }}
    </style>""",
        )

        return unicode2bytes(rendered_index, encoding='ascii')



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import abc
import base64
import os

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.error import ProcessExitedAlready
from twisted.logger import Logger
from twisted.python.failure import Failure

from buildbot import config
from buildbot.util import service
from buildbot.util.protocol import LineProcessProtocol

log = Logger()


# this is a BuildbotService, so that it can be started and destroyed.
# this is needed to implement kubectl proxy lifecycle
class KubeConfigLoaderBase(service.BuildbotService):
    name = "KubeConfig"

    @abc.abstractmethod
    def getConfig(self):
        """
        @return dictionary with optional params
        {
            'master_url': 'https://kube_master.url',
            'namespace': 'default_namespace',
            'headers' {
                'Authentication': XXX
            }
            # todo (quite hard to implement with treq):
            'cert': 'optional client certificate used to connect to ssl'
            'verify': 'kube master certificate authority to use to connect'
        }
        """

    def get_master_url(self):
        # This function may be called before reconfigService() is called.
        # The function must be overridden in case getConfig() is not fully setup in such situation.
        return self.getConfig()["master_url"]

    def getAuthorization(self):
        return None

    def __str__(self):
        """return unique str for SharedService"""
        # hash is implemented from ComparableMixin
        return f"{self.__class__.__name__}({hash(self)})"


class KubeHardcodedConfig(KubeConfigLoaderBase):
    def reconfigService(
        self,
        master_url=None,
        bearerToken=None,
        basicAuth=None,
        headers=None,
        cert=None,
        verify=None,
        namespace="default",
    ):
        self.config = {'master_url': master_url, 'namespace': namespace, 'headers': {}}
        if headers is not None:
            self.config['headers'] = headers
        if basicAuth and bearerToken:
            raise RuntimeError("set one of basicAuth and bearerToken, not both")
        self.basicAuth = basicAuth
        self.bearerToken = bearerToken
        if cert is not None:
            self.config['cert'] = cert
        if verify is not None:
            self.config['verify'] = verify

    checkConfig = reconfigService

    @defer.inlineCallbacks
    def getAuthorization(self):
        if self.basicAuth is not None:
            basicAuth = yield self.renderSecrets(self.basicAuth)
            authstring = f"{basicAuth['user']}:{basicAuth['password']}".encode()
            encoded = base64.b64encode(authstring)
            return f"Basic {encoded}"

        if self.bearerToken is not None:
            bearerToken = yield self.renderSecrets(self.bearerToken)
            return f"Bearer {bearerToken}"

        return None

    def getConfig(self):
        return self.config


class KubeCtlProxyConfigLoader(KubeConfigLoaderBase):
    """We use kubectl proxy to connect to kube master.
    Parsing the config and setting up SSL is complex.
    So for now, we use kubectl proxy to load the config and connect to master.
    This will run the kube proxy as a subprocess, and return configuration with
    http://localhost:PORT
    """

    kube_ctl_proxy_cmd = ['kubectl', 'proxy']  # for tests override

    class LocalPP(LineProcessProtocol):
        def __init__(self):
            super().__init__()
            self.got_output_deferred = defer.Deferred()
            self.terminated_deferred = defer.Deferred()
            self.first_line = b""

        def outLineReceived(self, line):
            if not self.got_output_deferred.called:
                self.got_output_deferred.callback(line)

        def errLineReceived(self, line):
            if not self.got_output_deferred.called:
                self.got_output_deferred.errback(Failure(RuntimeError(line)))

        def processEnded(self, status):
            super().processEnded(status)
            self.terminated_deferred.callback(None)

    def checkConfig(self, proxy_port=8001, namespace="default"):
        self.proxy_port = proxy_port
        self.namespace = namespace
        self.pp = None
        self.process = None

    @defer.inlineCallbacks
    def ensure_subprocess_killed(self):
        if self.pp is not None:
            try:
                self.process.signalProcess("TERM")
            except ProcessExitedAlready:
                pass  # oh well
            yield self.pp.terminated_deferred

    @defer.inlineCallbacks
    def reconfigService(self, proxy_port=8001, namespace="default"):
        self.proxy_port = proxy_port
        self.namespace = namespace

        if self.running:
            yield self.ensure_subprocess_killed()
            yield self.start_subprocess()

    @defer.inlineCallbacks
    def start_subprocess(self):
        self.pp = self.LocalPP()
        self.process = reactor.spawnProcess(
            self.pp,
            self.kube_ctl_proxy_cmd[0],
            [*self.kube_ctl_proxy_cmd, "-p", str(self.proxy_port)],
            env=os.environ,
        )
        self.kube_proxy_output = yield self.pp.got_output_deferred

    @defer.inlineCallbacks
    def startService(self):
        try:
            yield self.start_subprocess()
        except Exception:
            yield self.ensure_subprocess_killed()
            raise
        yield super().startService()

    @defer.inlineCallbacks
    def stopService(self):
        yield self.ensure_subprocess_killed()
        yield super().stopService()

    def getConfig(self):
        return {'master_url': f"http://localhost:{self.proxy_port}", 'namespace': self.namespace}


class KubeInClusterConfigLoader(KubeConfigLoaderBase):
    kube_dir = '/var/run/secrets/kubernetes.io/serviceaccount/'

    kube_namespace_file = os.path.join(kube_dir, 'namespace')
    kube_token_file = os.path.join(kube_dir, 'token')
    kube_cert_file = os.path.join(kube_dir, 'ca.crt')

    def checkConfig(self):
        if not os.path.exists(self.kube_dir):
            config.error(f"Not in kubernetes cluster (kube_dir not found: {self.kube_dir})")

    def reconfigService(self):
        self.config = {}
        self.config['master_url'] = self.get_master_url()
        self.config['verify'] = self.kube_cert_file
        with open(self.kube_token_file, encoding="utf-8") as token_content:
            token = token_content.read().strip()
            self.config['headers'] = {'Authorization': f'Bearer {token}'.format(token)}
        with open(self.kube_namespace_file, encoding="utf-8") as namespace_content:
            self.config['namespace'] = namespace_content.read().strip()

    def getConfig(self):
        return self.config

    def get_master_url(self):
        return os.environ["KUBERNETES_PORT"].replace("tcp", "https")


class KubeClientService(service.SharedService):
    name: str | None = "KubeClientService"  # type: ignore[assignment]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config_id_to_workers = {}
        self._worker_to_config = {}
        self._lock = defer.DeferredLock()

    @defer.inlineCallbacks
    def register(self, worker, config):
        yield self._lock.acquire()
        try:
            if worker.name in self._worker_to_config:
                raise ValueError(f"Worker {worker.name} registered multiple times")
            self._worker_to_config[worker.name] = config
            config_id = id(config)
            if config_id in self._config_id_to_workers:
                self._config_id_to_workers[config_id].append(worker.name)
            else:
                self._config_id_to_workers[config_id] = [worker.name]
                yield config.setServiceParent(self)
        finally:
            self._lock.release()

    @defer.inlineCallbacks
    def unregister(self, worker):
        yield self._lock.acquire()
        try:
            if worker.name not in self._worker_to_config:
                raise ValueError(f"Worker {worker.name} was not registered")
            config = self._worker_to_config.pop(worker.name)
            config_id = id(config)
            worker_list = self._config_id_to_workers[config_id]
            worker_list.remove(worker.name)
            if not worker_list:
                del self._config_id_to_workers[config_id]
                yield config.disownServiceParent()
        finally:
            self._lock.release()

    @defer.inlineCallbacks
    def startService(self):
        yield self._lock.acquire()
        try:
            yield super().startService()
        finally:
            self._lock.release()

    @defer.inlineCallbacks
    def stopService(self):
        yield self._lock.acquire()
        try:
            yield super().stopService()
        finally:
            self._lock.release()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
r"""
Buildbot metrics module

Keeps track of counts and timings of various internal buildbot
activities.

Basic architecture:

    MetricEvent.log(...)
          ||
          \/
    MetricLogObserver
          ||
          \/
    MetricHandler
          ||
          \/
    MetricWatcher
"""

from __future__ import annotations

import gc
import os
import sys
from collections import defaultdict
from collections import deque
from typing import TYPE_CHECKING

from twisted.application import service
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.python import log

from buildbot import util
from buildbot.util import service as util_service

if TYPE_CHECKING:
    from twisted.internet.base import ReactorBase

# Make use of the resource module if we can
try:
    import resource

    assert resource
except ImportError:
    resource = None  # type: ignore[assignment]


class MetricEvent:
    @classmethod
    def log(cls, *args, **kwargs):
        log.msg(metric=cls(*args, **kwargs))


class MetricCountEvent(MetricEvent):
    def __init__(self, counter, count=1, absolute=False):
        self.counter = counter
        self.count = count
        self.absolute = absolute


class MetricTimeEvent(MetricEvent):
    def __init__(self, timer, elapsed):
        self.timer = timer
        self.elapsed = elapsed


ALARM_OK, ALARM_WARN, ALARM_CRIT = list(range(3))
ALARM_TEXT = ["OK", "WARN", "CRIT"]


class MetricAlarmEvent(MetricEvent):
    def __init__(self, alarm, msg=None, level=ALARM_OK):
        self.alarm = alarm
        self.level = level
        self.msg = msg


def countMethod(counter):
    def decorator(func):
        def wrapper(*args, **kwargs):
            MetricCountEvent.log(counter=counter)
            return func(*args, **kwargs)

        return wrapper

    return decorator


class Timer:
    # For testing
    _reactor: ReactorBase | None = None

    def __init__(self, name):
        self.name = name
        self.started = None

    def startTimer(self, func):
        def wrapper(*args, **kwargs):
            self.start()
            return func(*args, **kwargs)

        return wrapper

    def stopTimer(self, func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            finally:
                self.stop()

        return wrapper

    def start(self):
        self.started = util.now(self._reactor)

    def stop(self):
        if self.started is not None:
            elapsed = util.now(self._reactor) - self.started
            MetricTimeEvent.log(timer=self.name, elapsed=elapsed)
            self.started = None


def timeMethod(name, _reactor=None):
    def decorator(func):
        t = Timer(name)
        t._reactor = _reactor

        def wrapper(*args, **kwargs):
            t.start()
            try:
                return func(*args, **kwargs)
            finally:
                t.stop()

        return wrapper

    return decorator


class FiniteList(deque):
    def __init__(self, maxlen=10):
        self._maxlen = maxlen
        super().__init__()

    def append(self, o):
        deque.append(self, o)
        if len(self) > self._maxlen:
            self.popleft()


class AveragingFiniteList(FiniteList):
    def __init__(self, maxlen=10):
        super().__init__(maxlen)
        self.average = 0

    def append(self, o):
        super().append(o)
        self._calc()

    def _calc(self):
        if not self:
            self.average = 0
        else:
            self.average = float(sum(self)) / len(self)

        return self.average


class MetricHandler:
    def __init__(self, metrics):
        self.metrics = metrics
        self.watchers = []

        self.reset()

    def addWatcher(self, watcher):
        self.watchers.append(watcher)

    def removeWatcher(self, watcher):
        self.watchers.remove(watcher)

    # For subclasses to define
    def reset(self):
        raise NotImplementedError

    def handle(self, eventDict, metric):
        raise NotImplementedError

    def get(self, metric):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    def report(self):
        raise NotImplementedError

    def asDict(self):
        raise NotImplementedError


class MetricCountHandler(MetricHandler):
    _counters: defaultdict[str, int] | None = None

    def reset(self):
        self._counters = defaultdict(int)

    def handle(self, eventDict, metric):
        if metric.absolute:
            self._counters[metric.counter] = metric.count
        else:
            self._counters[metric.counter] += metric.count

    def keys(self):
        return list(self._counters)

    def get(self, counter):
        return self._counters[counter]

    def report(self):
        retval = []
        for counter in sorted(self.keys()):
            retval.append(f"Counter {counter}: {self.get(counter)}")
        return "\n".join(retval)

    def asDict(self):
        retval = {}
        for counter in sorted(self.keys()):
            retval[counter] = self.get(counter)
        return {"counters": retval}


class MetricTimeHandler(MetricHandler):
    _timers: defaultdict[str, AveragingFiniteList] | None = None

    def reset(self):
        self._timers = defaultdict(AveragingFiniteList)

    def handle(self, eventDict, metric):
        self._timers[metric.timer].append(metric.elapsed)

    def keys(self):
        return list(self._timers)

    def get(self, timer):
        return self._timers[timer].average

    def report(self):
        retval = []
        for timer in sorted(self.keys()):
            retval.append(f"Timer {timer}: {self.get(timer):.3g}")
        return "\n".join(retval)

    def asDict(self):
        retval = {}
        for timer in sorted(self.keys()):
            retval[timer] = self.get(timer)
        return {"timers": retval}


class MetricAlarmHandler(MetricHandler):
    _alarms: defaultdict[str, tuple[int, str]] | None

    def reset(self):
        self._alarms = defaultdict(lambda x: ALARM_OK)

    def handle(self, eventDict, metric):
        self._alarms[metric.alarm] = (metric.level, metric.msg)

    def report(self):
        retval = []
        for alarm, (level, msg) in sorted(self._alarms.items()):
            if msg:
                retval.append(f"{ALARM_TEXT[level]} {alarm}: {msg}")
            else:
                retval.append(f"{ALARM_TEXT[level]} {alarm}")
        return "\n".join(retval)

    def asDict(self):
        retval = {}
        for alarm, (level, msg) in sorted(self._alarms.items()):
            retval[alarm] = (ALARM_TEXT[level], msg)
        return {"alarms": retval}


class AttachedWorkersWatcher:
    def __init__(self, metrics):
        self.metrics = metrics

    def run(self):
        # Check if 'BotMaster.attached_workers' equals
        # 'AbstractWorker.attached_workers'
        h = self.metrics.getHandler(MetricCountEvent)
        if not h:
            log.msg("Couldn't get MetricCountEvent handler")
            MetricAlarmEvent.log(
                'AttachedWorkersWatcher',
                msg="Coudln't get MetricCountEvent handler",
                level=ALARM_WARN,
            )
            return
        botmaster_count = h.get('BotMaster.attached_workers')
        worker_count = h.get('AbstractWorker.attached_workers')

        # We let these be off by one since they're counted at slightly
        # different times
        if abs(botmaster_count - worker_count) > 1:
            level = ALARM_WARN
        else:
            level = ALARM_OK

        MetricAlarmEvent.log(
            'attached_workers', msg=f'{botmaster_count} {worker_count}', level=level
        )


def _get_rss():
    if sys.platform == 'linux':
        try:
            with open(f"/proc/{os.getpid()}/statm", encoding='utf-8') as f:
                return int(f.read().split()[1])
        except Exception:
            return 0
    return 0


def periodicCheck(_reactor=reactor):
    try:
        # Measure how much garbage we have
        garbage_count = len(gc.garbage)
        MetricCountEvent.log('gc.garbage', garbage_count, absolute=True)
        if garbage_count == 0:
            level = ALARM_OK
        else:
            level = ALARM_WARN
        MetricAlarmEvent.log('gc.garbage', level=level)

        if resource:
            r = resource.getrusage(resource.RUSAGE_SELF)
            attrs = [
                'ru_utime',
                'ru_stime',
                'ru_maxrss',
                'ru_ixrss',
                'ru_idrss',
                'ru_isrss',
                'ru_minflt',
                'ru_majflt',
                'ru_nswap',
                'ru_inblock',
                'ru_oublock',
                'ru_msgsnd',
                'ru_msgrcv',
                'ru_nsignals',
                'ru_nvcsw',
                'ru_nivcsw',
            ]
            for i, a in enumerate(attrs):
                # Linux versions prior to 2.6.32 didn't report this value, but we
                # can calculate it from /proc/<pid>/statm
                v = r[i]
                if a == 'ru_maxrss' and v == 0:
                    v = _get_rss() * resource.getpagesize() / 1024
                MetricCountEvent.log(f'resource.{a}', v, absolute=True)
            MetricCountEvent.log('resource.pagesize', resource.getpagesize(), absolute=True)
        # Measure the reactor delay
        then = util.now(_reactor)
        dt = 0.1

        def cb():
            now = util.now(_reactor)
            delay = (now - then) - dt
            MetricTimeEvent.log("reactorDelay", delay)

        _reactor.callLater(dt, cb)
    except Exception:
        log.err(None, "while collecting VM metrics")


class MetricLogObserver(util_service.ReconfigurableServiceMixin, service.MultiService):
    _reactor = reactor

    def __init__(self):
        super().__init__()
        self.setName('metrics')

        self.enabled = False
        self.periodic_task = None
        self.periodic_interval = None
        self.log_task = None
        self.log_interval = None

        # Mapping of metric type to handlers for that type
        self.handlers = {}

        # Register our default handlers
        self.registerHandler(MetricCountEvent, MetricCountHandler(self))
        self.registerHandler(MetricTimeEvent, MetricTimeHandler(self))
        self.registerHandler(MetricAlarmEvent, MetricAlarmHandler(self))

        self.getHandler(MetricCountEvent).addWatcher(AttachedWorkersWatcher(self))

    def reconfigServiceWithBuildbotConfig(self, new_config):
        # first, enable or disable
        if new_config.metrics is None:
            self.disable()
        else:
            self.enable()

            metrics_config = new_config.metrics

            # Start up periodic logging
            log_interval = metrics_config.get('log_interval', 60)
            if log_interval != self.log_interval:
                if self.log_task:
                    self.log_task.stop()
                    self.log_task = None
                if log_interval:
                    self.log_task = LoopingCall(self.report)
                    self.log_task.clock = self._reactor
                    self.log_task.start(log_interval)

            # same for the periodic task
            periodic_interval = metrics_config.get('periodic_interval', 10)
            if periodic_interval != self.periodic_interval:
                if self.periodic_task:
                    self.periodic_task.stop()
                    self.periodic_task = None
                if periodic_interval:
                    self.periodic_task = LoopingCall(periodicCheck, self._reactor)
                    self.periodic_task.clock = self._reactor
                    self.periodic_task.start(periodic_interval)

        # upcall
        return super().reconfigServiceWithBuildbotConfig(new_config)

    def stopService(self):
        self.disable()
        super().stopService()

    def enable(self):
        if self.enabled:
            return
        log.addObserver(self.emit)
        self.enabled = True

    def disable(self):
        if not self.enabled:
            return

        if self.periodic_task:
            self.periodic_task.stop()
            self.periodic_task = None

        if self.log_task:
            self.log_task.stop()
            self.log_task = None

        log.removeObserver(self.emit)
        self.enabled = False

    def registerHandler(self, interface, handler):
        old = self.getHandler(interface)
        self.handlers[interface] = handler
        return old

    def getHandler(self, interface):
        return self.handlers.get(interface)

    def emit(self, eventDict):
        # Ignore non-statistic events
        metric = eventDict.get('metric')
        if not metric or not isinstance(metric, MetricEvent):
            return

        if metric.__class__ not in self.handlers:
            return

        h = self.handlers[metric.__class__]
        h.handle(eventDict, metric)
        for w in h.watchers:
            w.run()

    def asDict(self):
        retval = {}
        for _, handler in self.handlers.items():
            retval.update(handler.asDict())
        return retval

    def report(self):
        try:
            for handler in self.handlers.values():
                report = handler.report()
                if not report:
                    continue
                for line in report.split("\n"):
                    log.msg(line)
        except Exception:
            log.err(None, "generating metric report")



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import json as jsonmodule

from twisted.internet import defer
from twisted.logger import Logger
from twisted.python import deprecate
from twisted.python import versions
from zope.interface import implementer

from buildbot import util
from buildbot.interfaces import IHttpResponse
from buildbot.util import httpclientservice
from buildbot.util import service
from buildbot.util import toJson
from buildbot.util import unicode2bytes
from buildbot.util.twisted import async_to_deferred

log = Logger()


@implementer(IHttpResponse)
class ResponseWrapper:
    def __init__(self, code, content, url=None):
        self._content = content
        self._code = code
        self._url = url

    def content(self):
        content = unicode2bytes(self._content)
        return defer.succeed(content)

    def json(self):
        return defer.succeed(jsonmodule.loads(self._content))

    @property
    def code(self):
        return self._code

    @property
    def url(self):
        return self._url


class HTTPClientService(service.SharedService):
    """HTTPClientService is a SharedService class that fakes http requests for buildbot http
    service testing.

    This class is named the same as the real HTTPClientService so that it could replace the real
    class in tests. If a test creates this class earlier than the real one, fake is going to be
    used until the master is destroyed. Whenever a master wants to create real
    HTTPClientService, it will find an existing fake service with the same name and use it
    instead.
    """

    quiet = False

    def __init__(
        self, base_url, auth=None, headers=None, debug=False, verify=None, skipEncoding=False
    ):
        assert not base_url.endswith("/"), "baseurl should not end with /"
        super().__init__()
        self._session = httpclientservice.HTTPSession(
            self,
            base_url,
            auth=auth,
            headers=headers,
            debug=debug,
            verify=verify,
            skip_encoding=skipEncoding,
        )

        self._expected = []

    def updateHeaders(self, headers):
        self._session.update_headers(headers)

    @classmethod
    @async_to_deferred
    async def getService(cls, master, case, *args, **kwargs):
        def assertNotCalled(self, *_args, **_kwargs):
            case.fail(
                f"HTTPClientService called with *{_args!r}, **{_kwargs!r} "
                f"while should be called *{args!r} **{kwargs!r}"
            )

        case.patch(httpclientservice.HTTPClientService, "__init__", assertNotCalled)

        service = await super().getService(master, *args, **kwargs)
        service.case = case
        case.addCleanup(service.assertNoOutstanding)

        master.httpservice = service

        return service

    def expect(
        self,
        method,
        ep,
        session=None,
        params=None,
        headers=None,
        data=None,
        json=None,
        code=200,
        content=None,
        content_json=None,
        files=None,
        verify=None,
        cert=None,
        processing_delay_s=None,
    ):
        if content is not None and content_json is not None:
            return ValueError("content and content_json cannot be both specified")

        if content_json is not None:
            content = jsonmodule.dumps(content_json, default=toJson)

        self._expected.append({
            "method": method,
            "session": session,
            "ep": ep,
            "params": params,
            "headers": headers,
            "data": data,
            "json": json,
            "code": code,
            "content": content,
            "files": files,
            "verify": verify,
            "cert": cert,
            "processing_delay_s": processing_delay_s,
        })
        return None

    def assertNoOutstanding(self):
        self.case.assertEqual(
            0, len(self._expected), f"expected more http requests:\n {self._expected!r}"
        )

    @async_to_deferred
    async def _do_request(
        self,
        session,
        method,
        ep,
        params=None,
        headers=None,
        cookies=None,  # checks are not implemented
        data=None,
        json=None,
        files=None,
        auth=None,  # checks are not implemented
        timeout=None,
        verify=None,
        cert=None,
        allow_redirects=None,  # checks are not implemented
        proxies=None,  # checks are not implemented
    ) -> IHttpResponse:
        if ep.startswith('http://') or ep.startswith('https://'):
            pass
        else:
            assert ep == "" or ep.startswith("/"), "ep should start with /: " + ep

        if not self.quiet:
            log.debug(
                "{method} {ep} {params!r} <- {data!r}",
                method=method,
                ep=ep,
                params=params,
                data=data or json,
            )
        if json is not None:
            # ensure that the json is really jsonable
            jsonmodule.dumps(json, default=toJson)
        if files is not None:
            files = dict((k, v.read()) for (k, v) in files.items())
        if not self._expected:
            raise AssertionError(
                f"Not expecting a request, while we got: method={method!r}, ep={ep!r}, "
                f"params={params!r}, headers={headers!r}, data={data!r}, json={json!r}, "
                f"files={files!r}"
            )
        expect = self._expected.pop(0)
        processing_delay_s = expect.pop("processing_delay_s")

        expect_session = expect["session"] or self._session

        # pylint: disable=too-many-boolean-expressions
        if (
            expect_session.base_url != session.base_url
            or expect_session.auth != session.auth
            or expect_session.headers != session.headers
            or expect_session.verify != session.verify
            or expect_session.debug != session.debug
            or expect_session.skip_encoding != session.skip_encoding
            or expect["method"] != method
            or expect["ep"] != ep
            or expect["params"] != params
            or expect["headers"] != headers
            or expect["data"] != data
            or expect["json"] != json
            or expect["files"] != files
            or expect["verify"] != verify
            or expect["cert"] != cert
        ):
            raise AssertionError(
                "expecting:\n"
                f"session.base_url={expect_session.base_url!r}, "
                f"session.auth={expect_session.auth!r}, "
                f"session.headers={expect_session.headers!r}, "
                f"session.verify={expect_session.verify!r}, "
                f"session.debug={expect_session.debug!r}, "
                f"session.skip_encoding={expect_session.skip_encoding!r}, "
                f"method={expect['method']!r}, "
                f"ep={expect['ep']!r}, "
                f"params={expect['params']!r}, "
                f"headers={expect['headers']!r}, "
                f"data={expect['data']!r}, "
                f"json={expect['json']!r}, "
                f"files={expect['files']!r}, "
                f"verify={expect['verify']!r}, "
                f"cert={expect['cert']!r}"
                "\ngot      :\n"
                f"session.base_url={session.base_url!r}, "
                f"session.auth={session.auth!r}, "
                f"session.headers={session.headers!r}, "
                f"session.verify={session.verify!r}, "
                f"session.debug={session.debug!r}, "
                f"session.skip_encoding={session.skip_encoding!r}, "
                f"method={method!r}, "
                f"ep={ep!r}, "
                f"params={params!r}, "
                f"headers={headers!r}, "
                f"data={data!r}, "
                f"json={json!r}, "
                f"files={files!r}, "
                f"verify={verify!r}, "
                f"cert={cert!r}"
            )
        if not self.quiet:
            log.debug(
                "{method} {ep} -> {code} {content!r}",
                method=method,
                ep=ep,
                code=expect['code'],
                content=expect['content'],
            )

        if processing_delay_s is not None:
            await util.asyncSleep(1, reactor=self.master.reactor)

        return ResponseWrapper(expect['code'], expect['content'])

    # lets be nice to the auto completers, and don't generate that code
    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0))
    def get(self, ep, **kwargs):
        return self._do_request(self._session, 'get', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0))
    def put(self, ep, **kwargs):
        return self._do_request(self._session, 'put', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0))
    def delete(self, ep, **kwargs):
        return self._do_request(self._session, 'delete', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0))
    def post(self, ep, **kwargs):
        return self._do_request(self._session, 'post', ep, **kwargs)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from zope.interface import implementer

from buildbot import interfaces
from buildbot.process.build import Build
from buildbot.process.buildrequest import BuildRequest
from buildbot.process.properties import Properties
from buildbot.process.results import CANCELLED
from buildbot.reporters import utils
from buildbot.reporters.message import MessageFormatterRenderable

from .utils import BuildStatusGeneratorMixin


@implementer(interfaces.IReportGenerator)
class BuildRequestGenerator(BuildStatusGeneratorMixin):
    wanted_event_keys = [('buildrequests', None, 'new'), ('buildrequests', None, 'cancel')]

    compare_attrs: ClassVar[Sequence[str]] = ['formatter']

    def __init__(
        self,
        tags=None,
        builders=None,
        schedulers=None,
        branches=None,
        add_patch=False,
        formatter=None,
    ):
        super().__init__('all', tags, builders, schedulers, branches, None, None, add_patch)
        self.formatter = formatter
        if self.formatter is None:
            self.formatter = MessageFormatterRenderable('Build pending.')

    @defer.inlineCallbacks
    def partial_build_dict(self, master, buildrequest):
        brdict = yield master.db.buildrequests.getBuildRequest(buildrequest['buildrequestid'])
        bdict = {}

        props = Properties()
        buildrequest = yield BuildRequest.fromBrdict(master, brdict)
        builder = yield master.botmaster.getBuilderById(brdict.builderid)

        yield Build.setup_properties_known_before_build_starts(props, [buildrequest], builder)
        Build.setupBuildProperties(props, [buildrequest])

        bdict['properties'] = props.asDict()
        yield utils.get_details_for_buildrequest(master, brdict, bdict)
        return bdict

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, buildrequest):
        build = yield self.partial_build_dict(master, buildrequest)
        _, _, event = key
        if event == 'cancel':
            build['complete'] = True
            build['results'] = CANCELLED

        if not self.is_message_needed_by_props(build):
            return None

        report = yield self.buildrequest_message(master, build)
        return report

    @defer.inlineCallbacks
    def buildrequest_message(self, master, build):
        patches = self._get_patches_for_build(build)
        users = []
        buildmsg = yield self.formatter.format_message_for_build(
            master, build, is_buildset=True, mode=self.mode, users=users
        )

        return {
            'body': buildmsg['body'],
            'subject': buildmsg['subject'],
            'type': buildmsg['type'],
            'results': build['results'],
            'builds': [build],
            "buildset": build["buildset"],
            'users': list(users),
            'patches': patches,
            'logs': [],
            "extra_info": buildmsg["extra_info"],
        }



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import util


def patch_noargs_decorator(decorator):
    def new_decorator(func):
        wrapper = decorator(func)
        wrapper.__wrapped__ = func
        return wrapper

    util.mergeFunctionMetadata(decorator, new_decorator)
    return new_decorator


def patch():
    defer.inlineCallbacks = patch_noargs_decorator(defer.inlineCallbacks)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
import io
import os
import threading
from functools import partial
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer
from twisted.internet import threads
from twisted.python import log
from twisted.python import threadpool
from twisted.python.failure import Failure

from buildbot import util
from buildbot.config.master import get_is_in_unit_tests
from buildbot.db import base
from buildbot.db.compression import BrotliCompressor
from buildbot.db.compression import BZipCompressor
from buildbot.db.compression import CompressorInterface
from buildbot.db.compression import GZipCompressor
from buildbot.db.compression import LZ4Compressor
from buildbot.db.compression import ZStdCompressor
from buildbot.db.compression.protocol import CompressObjInterface
from buildbot.util.twisted import async_to_deferred
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from typing import AsyncGenerator
    from typing import Callable
    from typing import Generator
    from typing import Literal
    from typing import TypeVar

    from sqlalchemy.engine import Connection as SAConnection
    from twisted.internet.interfaces import IReactorThreads
    from typing_extensions import ParamSpec

    _T = TypeVar('_T')
    _P = ParamSpec('_P')

    LogType = Literal['s', 't', 'h', 'd']


class LogSlugExistsError(KeyError):
    pass


class LogCompressionFormatUnavailableError(LookupError):
    pass


@dataclasses.dataclass
class LogModel:
    id: int
    name: str
    slug: str
    stepid: int
    complete: bool
    num_lines: int
    type: LogType

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'LogsConnectorComponent '
                'getLog, getLogBySlug, and getLogs '
                'no longer return Log as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class RawCompressor(CompressorInterface):
    name = "raw"

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return data

    @staticmethod
    def read(data: bytes) -> bytes:
        return data

    class CompressObj(CompressObjInterface):
        def compress(self, data: bytes) -> bytes:
            return data

        def flush(self) -> bytes:
            return b''


class LogsConnectorComponent(base.DBConnectorComponent):
    # Postgres and MySQL will both allow bigger sizes than this.  The limit
    # for MySQL appears to be max_packet_size (default 1M).
    # note that MAX_CHUNK_SIZE is equal to BUFFER_SIZE in buildbot_worker.runprocess
    MAX_CHUNK_SIZE = 65536  # a chunk may not be bigger than this
    MAX_CHUNK_LINES = 1000  # a chunk may not have more lines than this

    NO_COMPRESSION_ID = 0
    COMPRESSION_BYID: dict[int, type[CompressorInterface]] = {
        NO_COMPRESSION_ID: RawCompressor,
        1: GZipCompressor,
        2: BZipCompressor,
        3: LZ4Compressor,
        4: ZStdCompressor,
        5: BrotliCompressor,
    }

    COMPRESSION_MODE = {
        compressor.name: (compressor_id, compressor)
        for compressor_id, compressor in COMPRESSION_BYID.items()
    }

    def __init__(self, connector: base.DBConnector):
        super().__init__(connector)

        max_threads = 1
        if cpu_count := os.cpu_count():
            # use at most half cpu available to avoid oversubscribing
            # the master on other processes
            max_threads = max(int(cpu_count / 2), max_threads)

        self._compression_pool = util.twisted.ThreadPool(
            minthreads=1,
            maxthreads=max_threads,
            name='DBLogCompression',
        )

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        self._compression_pool.start()

    @defer.inlineCallbacks
    def stopService(self):
        yield super().stopService()
        self._compression_pool.stop()

    def _defer_to_compression_pool(
        self,
        callable: Callable[_P, _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> defer.Deferred[_T]:
        return threads.deferToThreadPool(
            self.master.reactor, self._compression_pool, callable, *args, **kwargs
        )

    def _get_compressor(self, compressor_id: int) -> type[CompressorInterface]:
        compressor = self.COMPRESSION_BYID.get(compressor_id)
        if compressor is None:
            msg = f"Unknown compression method ID {compressor_id}"
            raise LogCompressionFormatUnavailableError(msg)
        if not compressor.available:
            msg = (
                f"Log compression method {compressor.name} is not available. "
                "You might be missing a dependency."
            )
            raise LogCompressionFormatUnavailableError(msg)
        return compressor

    def _getLog(self, whereclause) -> defer.Deferred[LogModel | None]:
        def thd_getLog(conn) -> LogModel | None:
            q = self.db.model.logs.select()
            if whereclause is not None:
                q = q.where(whereclause)
            res = conn.execute(q).mappings()
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return self.db.pool.do(thd_getLog)

    def getLog(self, logid: int) -> defer.Deferred[LogModel | None]:
        return self._getLog(self.db.model.logs.c.id == logid)

    def getLogBySlug(self, stepid: int, slug: str) -> defer.Deferred[LogModel | None]:
        tbl = self.db.model.logs
        return self._getLog((tbl.c.slug == slug) & (tbl.c.stepid == stepid))

    def getLogs(self, stepid: int | None = None) -> defer.Deferred[list[LogModel]]:
        def thdGetLogs(conn) -> list[LogModel]:
            tbl = self.db.model.logs
            q = tbl.select()
            if stepid is not None:
                q = q.where(tbl.c.stepid == stepid)
            q = q.order_by(tbl.c.id)
            res = conn.execute(q).mappings()
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thdGetLogs)

    async def iter_log_lines(
        self,
        logid: int,
        first_line: int = 0,
        last_line: int | None = None,
    ) -> AsyncGenerator[str, None]:
        def _thd_get_chunks(
            conn: SAConnection,
            first_line: int,
            last_line: int | None,
            batch: int,
        ) -> list[tuple[int, int, int, bytes]]:
            tbl = self.db.model.logchunks
            q = sa.select(tbl.c.first_line, tbl.c.last_line, tbl.c.content, tbl.c.compressed)
            q = q.where(tbl.c.logid == logid)
            if last_line is not None:
                q = q.where(tbl.c.first_line <= last_line)
            q = q.where(tbl.c.last_line >= first_line)

            q = q.order_by(tbl.c.first_line)
            if batch > 0:
                q = q.limit(batch)

            return [
                (row.first_line, row.last_line, row.compressed, row.content)
                for row in conn.execute(q)
            ]

        async def _iter_chunks_batched():
            CHUNK_BATCH_SIZE = 100
            batch_first_line = first_line
            while chunks := await self.db.pool.do(
                _thd_get_chunks,
                batch_first_line,
                last_line,
                CHUNK_BATCH_SIZE,
            ):
                for chunk in chunks:
                    yield chunk

                _, chunk_last_line, _, _ = chunks[-1]
                batch_first_line = max(batch_first_line, chunk_last_line) + 1

        def _iter_uncompress_lines(
            chunk_first_line: int,
            compressed: int,
            content: bytes,
        ) -> Generator[str, None, None]:
            # Retrieve associated "reader" and extract the data
            # Note that row.content is stored as bytes, and our caller expects unicode
            data = self._get_compressor(compressed).read(content)
            # NOTE: we need a streaming decompression interface
            with io.BytesIO(data) as data_buffer, io.TextIOWrapper(
                data_buffer,
                encoding='utf-8',
            ) as reader:
                # last line-ending is stripped from chunk on insert
                # add it back here to simplify handling after
                data_buffer.seek(0, os.SEEK_END)
                data_buffer.write(b'\n')
                data_buffer.seek(0, os.SEEK_SET)

                line_idx = chunk_first_line

                # need to skip some lines
                while line_idx < first_line and reader.readline():
                    line_idx += 1

                while (last_line is None or line_idx <= last_line) and (line := reader.readline()):
                    yield line
                    line_idx += 1

        async for chunk_first_line, _, compressed, content in _iter_chunks_batched():
            async for line in _async_iter_on_pool(
                partial(
                    _iter_uncompress_lines,
                    chunk_first_line=chunk_first_line,
                    compressed=compressed,
                    content=content,
                ),
                reactor=self.master.reactor,
                provider_threadpool=self._compression_pool,
                # disable limit as we process one chunk at a time
                max_backlog=0,
            ):
                yield line

    @async_to_deferred
    async def getLogLines(self, logid: int, first_line: int, last_line: int) -> str:
        lines: list[str] = []
        async for line in self.iter_log_lines(
            logid=logid,
            first_line=first_line,
            last_line=last_line,
        ):
            lines.append(line)

        return ''.join(lines)

    def addLog(self, stepid: int, name: str, slug: str, type: LogType) -> defer.Deferred[int]:
        assert type in 'tsh', "Log type must be one of t, s, or h"

        def thdAddLog(conn) -> int:
            try:
                r = conn.execute(
                    self.db.model.logs.insert(),
                    {
                        "name": name,
                        "slug": slug,
                        "stepid": stepid,
                        "complete": 0,
                        "num_lines": 0,
                        "type": type,
                    },
                )
                conn.commit()
                return r.inserted_primary_key[0]
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError) as e:
                conn.rollback()
                raise LogSlugExistsError(
                    f"log with slug '{slug!r}' already exists in this step"
                ) from e

        return self.db.pool.do(thdAddLog)

    def _get_configured_compressor(self) -> tuple[int, type[CompressorInterface]]:
        compress_method: str = self.master.config.logCompressionMethod
        return self.COMPRESSION_MODE.get(compress_method, (self.NO_COMPRESSION_ID, RawCompressor))

    @async_to_deferred
    async def appendLog(self, logid: int, content: str) -> tuple[int, int] | None:
        def _thd_get_numlines(conn: SAConnection) -> int | None:
            q = sa.select(self.db.model.logs.c.num_lines)
            q = q.where(self.db.model.logs.c.id == logid)
            res = conn.execute(q)
            num_lines = res.fetchone()
            res.close()
            return num_lines[0] if num_lines else None

        def _thd_insert_chunk(
            conn: SAConnection,
            first_line: int,
            last_line: int,
            content: bytes,
            compressed_id: int,
        ) -> None:
            res = conn.execute(
                self.db.model.logchunks.insert(),
                {
                    "logid": logid,
                    "first_line": first_line,
                    "last_line": last_line,
                    "content": content,
                    "compressed": compressed_id,
                },
            )
            conn.commit()
            res.close()

        def _thd_update_num_lines(conn: SAConnection, num_lines: int) -> None:
            res = conn.execute(
                self.db.model.logs.update()
                .where(self.db.model.logs.c.id == logid)
                .values(num_lines=num_lines)
            )
            conn.commit()
            res.close()

        def _thd_compress_chunk(
            compress_obj: CompressObjInterface,
            compressor_id: int,
            lines: list[bytes],
        ) -> tuple[bytes, int, int]:
            # check for trailing newline and strip it for storage
            # chunks omit the trailing newline
            assert lines and lines[-1][-1:] == b'\n'
            lines[-1] = lines[-1][:-1]

            compressed_bytes: list[bytes] = []
            uncompressed_size = 0
            for line in lines:
                uncompressed_size += len(line)
                compressed_bytes.append(compress_obj.compress(line))
            compressed_bytes.append(compress_obj.flush())
            compressed_chunk = b''.join(compressed_bytes)

            # Is it useful to compress the chunk?
            if uncompressed_size <= len(compressed_chunk):
                return b''.join(lines), self.NO_COMPRESSION_ID, len(lines)

            return compressed_chunk, compressor_id, len(lines)

        def _thd_iter_chunk_compress(
            content: str,
        ) -> Generator[tuple[bytes, int, int], None]:
            """
            Split content into chunk delimited by line-endings.
            Try our best to keep chunks smaller than MAX_CHUNK_SIZE
            """

            def _truncate_line(line: bytes) -> bytes:
                log.msg(f'truncating long line for log {logid}')
                line = line[: self.MAX_CHUNK_SIZE - 1]
                while line:
                    try:
                        line.decode('utf-8')
                        break
                    except UnicodeDecodeError:
                        line = line[:-1]
                return line + b'\n'

            compressor_id, compressor = self._get_configured_compressor()
            compress_obj = compressor.CompressObj()

            with io.StringIO(content) as buffer:
                lines: list[bytes] = []
                lines_size = 0
                while line := buffer.readline():
                    line_bytes = line.encode('utf-8')
                    line_size = len(line_bytes)
                    # would this go over limit?
                    if lines and lines_size + line_size > self.MAX_CHUNK_SIZE:
                        # flush lines
                        yield _thd_compress_chunk(compress_obj, compressor_id, lines)
                        del lines[:]
                        lines_size = 0

                    if line_size > self.MAX_CHUNK_SIZE:
                        compressed = _thd_compress_chunk(compress_obj, compressor_id, [line_bytes])
                        compressed_chunk, _, _ = compressed
                        # check if compressed size is compliant with DB row limit
                        if len(compressed_chunk) > self.MAX_CHUNK_SIZE:
                            compressed = _thd_compress_chunk(
                                compress_obj, compressor_id, [_truncate_line(line_bytes)]
                            )
                        yield compressed
                    else:
                        lines.append(line_bytes)

                if lines:
                    yield _thd_compress_chunk(compress_obj, compressor_id, lines)

        assert content[-1] == '\n'

        num_lines = await self.db.pool.do(_thd_get_numlines)
        if num_lines is None:
            # ignore a missing log
            return None

        # Break the content up into chunks
        chunk_first_line = last_line = num_lines
        async for (
            compressed_chunk,
            compressed_id,
            chunk_lines_count,
        ) in _async_iter_on_pool(
            partial(
                _thd_iter_chunk_compress,
                content=content,
            ),
            reactor=self.master.reactor,
            provider_threadpool=self._compression_pool,
            # In theory, memory usage could grow to:
            # MAX_CHUNK_SIZE * max_backlog PER thread (capped by _compression_pool.maxthreads)
            # with:
            #   - MAX_CHUNK_SIZE = 64KB
            #   - max_backlog = 100
            # ~6MB per thread
            max_backlog=100,
        ):
            last_line = chunk_first_line + chunk_lines_count - 1

            await self.db.pool.do(
                _thd_insert_chunk,
                first_line=chunk_first_line,
                last_line=last_line,
                content=compressed_chunk,
                compressed_id=compressed_id,
            )

            chunk_first_line = last_line + 1

        await self.db.pool.do(_thd_update_num_lines, last_line + 1)
        return num_lines, last_line

    def finishLog(self, logid: int) -> defer.Deferred[None]:
        def thdfinishLog(conn) -> None:
            tbl = self.db.model.logs
            q = tbl.update().where(tbl.c.id == logid)
            conn.execute(q.values(complete=1))

        return self.db.pool.do_with_transaction(thdfinishLog)

    @async_to_deferred
    async def compressLog(self, logid: int, force: bool = False) -> int:
        """
        returns the size (in bytes) saved.
        """
        tbl = self.db.model.logchunks

        def _thd_gather_chunks_to_process(conn: SAConnection) -> list[tuple[int, int]]:
            """
            returns the total size of chunks and a list of chunks to group.
            chunks list is empty if not force, and no chunks would be grouped.
            """
            q = (
                sa.select(
                    tbl.c.first_line,
                    tbl.c.last_line,
                    sa.func.length(tbl.c.content),
                )
                .where(tbl.c.logid == logid)
                .order_by(tbl.c.first_line)
            )

            rows = conn.execute(q)

            # get the first chunk to seed new_chunks list
            first_chunk = next(rows, None)
            if first_chunk is None:
                # no chunks in log, early out
                return []

            grouped_chunks: list[tuple[int, int]] = [
                (first_chunk.first_line, first_chunk.last_line)
            ]

            # keep track of how many chunks we use now
            # to compare with grouped chunks and
            # see if we need to do some work
            # start at 1 since we already queries one above
            current_chunk_count = 1

            current_group_new_size = first_chunk.length_1
            # first pass, we fetch the full list of chunks (without content) and find out
            # the chunk groups which could use some gathering.
            for row in rows:
                current_chunk_count += 1

                chunk_first_line: int = row.first_line
                chunk_last_line: int = row.last_line
                chunk_size: int = row.length_1

                group_first_line, _group_last_line = grouped_chunks[-1]

                can_merge_chunks = (
                    # note that we count the compressed size for efficiency reason
                    # unlike to the on-the-flow chunk splitter
                    current_group_new_size + chunk_size <= self.MAX_CHUNK_SIZE
                    and (chunk_last_line - group_first_line) <= self.MAX_CHUNK_LINES
                )
                if can_merge_chunks:
                    # merge chunks, since we ordered the query by 'first_line'
                    # and we assume that chunks are contiguous, it's pretty easy
                    grouped_chunks[-1] = (group_first_line, chunk_last_line)
                    current_group_new_size += chunk_size
                else:
                    grouped_chunks.append((chunk_first_line, chunk_last_line))
                    current_group_new_size = chunk_size

            rows.close()

            if not force and current_chunk_count <= len(grouped_chunks):
                return []

            return grouped_chunks

        def _thd_get_chunks_content(
            conn: SAConnection,
            first_line: int,
            last_line: int,
        ) -> list[tuple[int, bytes]]:
            q = (
                sa.select(tbl.c.content, tbl.c.compressed)
                .where(tbl.c.logid == logid)
                .where(tbl.c.first_line >= first_line)
                .where(tbl.c.last_line <= last_line)
                .order_by(tbl.c.first_line)
            )
            rows = conn.execute(q)
            content = [(row.compressed, row.content) for row in rows]
            rows.close()
            return content

        def _thd_replace_chunks_by_new_grouped_chunk(
            conn: SAConnection,
            first_line: int,
            last_line: int,
            new_compressed_id: int,
            new_content: bytes,
        ) -> None:
            # Transaction is necessary so that readers don't see disappeared chunks
            with conn.begin():
                # we remove the chunks that we are compressing
                deletion_query = (
                    tbl.delete()
                    .where(tbl.c.logid == logid)
                    .where(tbl.c.first_line >= first_line)
                    .where(tbl.c.last_line <= last_line)
                )
                conn.execute(deletion_query).close()

                # and we recompress them in one big chunk
                conn.execute(
                    tbl.insert(),
                    {
                        "logid": logid,
                        "first_line": first_line,
                        "last_line": last_line,
                        "content": new_content,
                        "compressed": new_compressed_id,
                    },
                ).close()

                conn.commit()

        def _thd_recompress_chunks(
            compressed_chunks: list[tuple[int, bytes]],
            compress_obj: CompressObjInterface,
        ) -> tuple[bytes, int]:
            """This has to run in the compression thread pool"""
            # decompress this group of chunks. Note that the content is binary bytes.
            # no need to decode anything as we are going to put in back stored as bytes anyway
            chunks: list[bytes] = []
            bytes_saved = 0
            for idx, (chunk_compress_id, chunk_content) in enumerate(compressed_chunks):
                bytes_saved += len(chunk_content)

                # trailing line-ending is stripped from chunks
                # need to add it back, except for the last one
                if idx != 0:
                    chunks.append(compress_obj.compress(b'\n'))

                uncompressed_content = self._get_compressor(chunk_compress_id).read(chunk_content)
                chunks.append(compress_obj.compress(uncompressed_content))

            chunks.append(compress_obj.flush())
            new_content = b''.join(chunks)
            bytes_saved -= len(new_content)
            return new_content, bytes_saved

        chunk_groups = await self.db.pool.do(_thd_gather_chunks_to_process)
        if not chunk_groups:
            return 0

        total_bytes_saved: int = 0

        compressed_id, compressor = self._get_configured_compressor()
        compress_obj = compressor.CompressObj()
        for group_first_line, group_last_line in chunk_groups:
            compressed_chunks = await self.db.pool.do(
                _thd_get_chunks_content,
                first_line=group_first_line,
                last_line=group_last_line,
            )

            new_content, bytes_saved = await self._defer_to_compression_pool(
                _thd_recompress_chunks,
                compressed_chunks=compressed_chunks,
                compress_obj=compress_obj,
            )

            total_bytes_saved += bytes_saved

            await self.db.pool.do(
                _thd_replace_chunks_by_new_grouped_chunk,
                first_line=group_first_line,
                last_line=group_last_line,
                new_compressed_id=compressed_id,
                new_content=new_content,
            )

        return total_bytes_saved

    def deleteOldLogChunks(self, older_than_timestamp: int) -> defer.Deferred[int]:
        def thddeleteOldLogs(conn) -> int:
            model = self.db.model
            res = conn.execute(sa.select(sa.func.count(model.logchunks.c.logid)))
            count1 = res.fetchone()[0]
            res.close()

            # update log types older than timestamps
            # we do it first to avoid having UI discrepancy

            # N.B.: we utilize the fact that steps.id is auto-increment, thus steps.started_at
            # times are effectively sorted and we only need to find the steps.id at the upper
            # bound of steps to update.

            # SELECT steps.id from steps WHERE steps.started_at < older_than_timestamp ORDER BY
            # steps.id DESC LIMIT 1;
            res = conn.execute(
                sa.select(model.steps.c.id)
                .where(model.steps.c.started_at < older_than_timestamp)
                .order_by(model.steps.c.id.desc())
                .limit(1)
            )
            res_list = res.fetchone()
            stepid_max = None
            if res_list:
                stepid_max = res_list[0]
            res.close()

            # UPDATE logs SET logs.type = 'd' WHERE logs.stepid <= stepid_max AND type != 'd';
            if stepid_max:
                res = conn.execute(
                    model.logs.update()
                    .where(sa.and_(model.logs.c.stepid <= stepid_max, model.logs.c.type != 'd'))
                    .values(type='d')
                )
                conn.commit()
                res.close()

            # query all logs with type 'd' and delete their chunks.
            if self.db._engine.dialect.name == 'sqlite':
                # sqlite does not support delete with a join, so for this case we use a subquery,
                # which is much slower
                q = sa.select(model.logs.c.id)
                q = q.select_from(model.logs)
                q = q.where(model.logs.c.type == 'd')

                # delete their logchunks
                q = model.logchunks.delete().where(model.logchunks.c.logid.in_(q))
            else:
                q = model.logchunks.delete()
                q = q.where(model.logs.c.id == model.logchunks.c.logid)
                q = q.where(model.logs.c.type == 'd')

            res = conn.execute(q)
            conn.commit()
            res.close()
            res = conn.execute(sa.select(sa.func.count(model.logchunks.c.logid)))
            count2 = res.fetchone()[0]
            res.close()
            return count1 - count2

        return self.db.pool.do(thddeleteOldLogs)

    def _model_from_row(self, row):
        return LogModel(
            id=row.id,
            name=row.name,
            slug=row.slug,
            stepid=row.stepid,
            complete=bool(row.complete),
            num_lines=row.num_lines,
            type=row.type,
        )


async def _async_iter_on_pool(
    generator_sync: Callable[[], Generator[_T, None, None]],
    *,
    reactor: IReactorThreads,
    provider_threadpool: threadpool.ThreadPool | None = None,
    max_backlog: int = 1,
    wait_backlog_consuption: bool = True,
) -> AsyncGenerator[_T, None]:
    """
    Utility to transform a sync `Generator` function into an `AsyncGenerator`
    by executing it on a threadpool.

    :param generator_sync:
        sync Generator function (if arguments are necessary, use functools.partial)

    :param reactor: Twisted reactor to use

    :param provider_threadpool:
        Threadpool to run the Generator on (default to reactor's ThreadPool)

    :param max_backlog:
        Maximum size of the buffer used to communicate between sync and async Generators.

        A value of 0 or less means unlimited.

        When the buffer contains `max_backlog` items,
        the threaded sync Generator will wait until at least one element is consumed.

        Note: this is forced to `0` if in unit tests and `provider_threadpool` is a `NonThreadPool`.

    :param wait_backlog_consuption:
        If `True`, will wait until all items in the buffer are consumed.

        This is used to prevent a new threadpool task to run,
        potentially creating a new buffer consuming memory
        while the previous buffer is still in use.

        Note: this is forced to `False` if in unit tests and `provider_threadpool` is a `NonThreadPool`.
    """

    if provider_threadpool is None:
        provider_threadpool = reactor.getThreadPool()

    # create a single element queue as to
    # occupy a thread of the pool
    # avoiding too many compressed chunks in memory awaiting DB insert
    # use 0 (unlimited) in tests as there isn't really a threadpool / reactor running
    if get_is_in_unit_tests():
        from buildbot.test.fake.reactor import NonThreadPool

        if isinstance(provider_threadpool, NonThreadPool):
            max_backlog = 0
            wait_backlog_consuption = False

    queue: defer.DeferredQueue[_T | _CloseObj] = defer.DeferredQueue()

    condition = threading.Condition()

    # dummy object that resolve the callback of the task.
    # Needed as we can't know what the callable will provide,
    # so None, False, ... can't be used.
    # But, we know that callback will return None, so we can
    # override it's callback result
    class _CloseObj:
        pass

    close_obj = _CloseObj()

    def _can_put_in_queue():
        return max_backlog <= 0 or len(queue.pending) < max_backlog

    def _provider_wrapped() -> None:
        try:
            for item in generator_sync():
                with condition:
                    condition.wait_for(_can_put_in_queue)
                reactor.callFromThread(queue.put, item)
        finally:
            if wait_backlog_consuption:
                with condition:
                    condition.wait_for(lambda: len(queue.pending) <= 0)

    def _put_close(res: None | Failure) -> None | Failure:
        queue.put(close_obj)
        return res

    worker_task = threads.deferToThreadPool(
        reactor,
        provider_threadpool,
        _provider_wrapped,
    ).addBoth(callback=_put_close)

    while (item := await queue.get()) is not close_obj:
        assert not isinstance(item, _CloseObj)
        with condition:
            condition.notify()
        yield item

    assert worker_task.called
    # so that if task ended in exception, it's correctly propagated
    # but only if error, as await a successfully resolved Deferred will
    # never finish
    if isinstance(worker_task.result, Failure):
        await worker_task



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from buildbot.db.build_data import BuildDataModel


def _db2data(model: BuildDataModel):
    return {
        'buildid': model.buildid,
        'name': model.name,
        'value': model.value,
        'length': model.length,
        'source': model.source,
    }


class BuildDatasNoValueEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/builders/n:builderid/builds/n:build_number/data",
        "/builders/s:buildername/builds/n:build_number/data",
        "/builds/n:buildid/data",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        buildid = yield self.getBuildid(kwargs)

        build_datadicts = yield self.master.db.build_data.getAllBuildDataNoValues(buildid)

        results = []
        for dbdict in build_datadicts:
            results.append(_db2data(dbdict))
        return results


class BuildDataNoValueEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/builders/n:builderid/builds/n:build_number/data/i:name",
        "/builders/s:buildername/builds/n:build_number/data/i:name",
        "/builds/n:buildid/data/i:name",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        buildid = yield self.getBuildid(kwargs)
        name = kwargs['name']

        build_datadict = yield self.master.db.build_data.getBuildDataNoValue(buildid, name)

        return _db2data(build_datadict) if build_datadict else None


class BuildDataEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.RAW
    pathPatterns = [
        "/builders/n:builderid/builds/n:build_number/data/i:name/value",
        "/builders/s:buildername/builds/n:build_number/data/i:name/value",
        "/builds/n:buildid/data/i:name/value",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        buildid = yield self.getBuildid(kwargs)
        name = kwargs['name']

        dbdict = yield self.master.db.build_data.getBuildData(buildid, name)
        if not dbdict:
            return None

        return {
            'raw': dbdict.value,
            'mime-type': 'application/octet-stream',
            'filename': dbdict.name,
        }


class BuildData(base.ResourceType):
    name = "build_data"
    plural = "build_data"
    endpoints = [BuildDatasNoValueEndpoint, BuildDataNoValueEndpoint, BuildDataEndpoint]

    class EntityType(types.Entity):
        buildid = types.Integer()
        name = types.String()
        length = types.Integer()
        value = types.NoneOk(types.Binary())
        source = types.String()

    entityType = EntityType(name)

    @base.updateMethod
    def setBuildData(
        self, buildid: int, name: str, value: bytes, source: str
    ) -> defer.Deferred[None]:
        # forward deferred directly
        return self.master.db.build_data.setBuildData(buildid, name, value, source)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.python import log

from buildbot.data import base

if TYPE_CHECKING:
    from typing import Sequence


class NotSupportedFieldTypeError(TypeError):
    def __init__(self, data, *args: object) -> None:
        super().__init__(
            (f"Unsupported data type '{type(data)}': must be an instance of Dict or a Dataclass."),
            *args,
        )


def _data_getter(d, fld):
    if isinstance(d, dict):
        return d[fld]
    if dataclasses.is_dataclass(d):
        try:
            return getattr(d, fld)
        except AttributeError as e:
            # backward compatibility when only dict was allowed
            raise KeyError(*e.args) from e

    raise NotSupportedFieldTypeError(d)


class FieldBase:
    """
    This class implements a basic behavior
    to wrap value into a `Field` instance

    """

    __slots__ = ['field', 'op', 'values']

    singular_operators = {
        'eq': lambda d, v: d == v[0],
        'ne': lambda d, v: d != v[0],
        'lt': lambda d, v: d < v[0],
        'le': lambda d, v: d <= v[0],
        'gt': lambda d, v: d > v[0],
        'ge': lambda d, v: d >= v[0],
        'contains': lambda d, v: v[0] in d,
        'in': lambda d, v: d in v,
        'notin': lambda d, v: d not in v,
    }

    singular_operators_sql = {
        'eq': lambda d, v: d == v[0],
        'ne': lambda d, v: d != v[0],
        'lt': lambda d, v: d < v[0],
        'le': lambda d, v: d <= v[0],
        'gt': lambda d, v: d > v[0],
        'ge': lambda d, v: d >= v[0],
        'contains': lambda d, v: d.contains(v[0]),
        # only support string values, because currently there are no queries against lists in SQL
        'in': lambda d, v: d.in_(v),
        'notin': lambda d, v: d.notin_(v),
    }

    plural_operators = {
        'eq': lambda d, v: d in v,
        'ne': lambda d, v: d not in v,
        'contains': lambda d, v: len(set(v).intersection(set(d))) > 0,
        'in': lambda d, v: d in v,
        'notin': lambda d, v: d not in v,
    }

    plural_operators_sql = {
        'eq': lambda d, v: d.in_(v),
        'ne': lambda d, v: d.notin_(v),
        'contains': lambda d, vs: sa.or_(*[d.contains(v) for v in vs]),
        'in': lambda d, v: d.in_(v),
        'notin': lambda d, v: d.notin_(v),
        # sqlalchemy v0.8's or_ cannot take generator arguments, so this has to be manually expanded
        # only support string values, because currently there are no queries against lists in SQL
    }

    def __init__(self, field: bytes | str, op: str, values: Sequence | set):
        self.field = field
        self.op = op
        self.values = values
        # `str` is a Sequence as well...
        assert not isinstance(values, str)

    def getOperator(self, sqlMode=False):
        v = self.values
        if len(v) == 1:
            if sqlMode:
                ops = self.singular_operators_sql
            else:
                ops = self.singular_operators
        else:
            if sqlMode:
                ops = self.plural_operators_sql
            else:
                ops = self.plural_operators
            v = set(v)
        return ops[self.op]

    def apply(self, data):
        fld = self.field
        v = self.values
        f = self.getOperator()
        return (d for d in data if f(_data_getter(d, fld), v))

    def __repr__(self):
        return f"resultspec.{self.__class__.__name__}('{self.field}','{self.op}',{self.values})"

    def __eq__(self, b):
        for i in self.__slots__:
            if getattr(self, i) != getattr(b, i):
                return False
        return True

    def __ne__(self, b):
        return not self == b


class Property(FieldBase):
    """
    Wraps ``property`` type value(s)

    """


class Filter(FieldBase):
    """
    Wraps ``filter`` type value(s)

    """


class NoneComparator:
    """
    Object which wraps 'None' when doing comparisons in sorted().
    '> None' and '< None' are not supported
    in Python 3.
    """

    def __init__(self, value):
        self.value = value

    def __lt__(self, other):
        if self.value is None and other.value is None:
            return False
        elif self.value is None:
            return True
        elif other.value is None:
            return False
        return self.value < other.value

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __gt__(self, other):
        if self.value is None and other.value is None:
            return False
        elif self.value is None:
            return False
        elif other.value is None:
            return True
        return self.value > other.value


class ReverseComparator:
    """
    Object which swaps '<' and '>' so
    instead of a < b, it does b < a,
    and instead of a > b, it does b > a.
    This can be used in reverse comparisons.
    """

    def __init__(self, value):
        self.value = value

    def __lt__(self, other):
        return other.value < self.value

    def __eq__(self, other):
        return other.value == self.value

    def __ne__(self, other):
        return other.value != self.value

    def __gt__(self, other):
        return other.value > self.value


class ResultSpec:
    __slots__ = ['fieldMapping', 'fields', 'filters', 'limit', 'offset', 'order', 'properties']

    def __init__(
        self, filters=None, fields=None, properties=None, order=None, limit=None, offset=None
    ):
        self.filters = filters or []
        self.properties = properties or []
        self.fields = fields
        self.order = order
        self.limit = limit
        self.offset = offset
        self.fieldMapping = {}

    def __repr__(self):
        return (
            f"ResultSpec(**{{'filters': {self.filters}, 'fields': {self.fields}, "
            f"'properties': {self.properties}, 'order': {self.order}, 'limit': {self.limit}, "
            f"'offset': {self.offset}" + "})"
        )

    def __eq__(self, b):
        for i in ['filters', 'fields', 'properties', 'order', 'limit', 'offset']:
            if getattr(self, i) != getattr(b, i):
                return False
        return True

    def __ne__(self, b):
        return not self == b

    def popProperties(self):
        values = []
        for p in self.properties:
            if p.field == b'property' and p.op == 'eq':
                self.properties.remove(p)
                values = p.values
                break
        return values

    def popFilter(self, field, op):
        for f in self.filters:
            if f.field == field and f.op == op:
                self.filters.remove(f)
                return f.values
        return None

    def popOneFilter(self, field, op):
        v = self.popFilter(field, op)
        return v[0] if v is not None else None

    def popBooleanFilter(self, field):
        eqVals = self.popFilter(field, 'eq')
        if eqVals and len(eqVals) == 1:
            return eqVals[0]
        neVals = self.popFilter(field, 'ne')
        if neVals and len(neVals) == 1:
            return not neVals[0]
        return None

    def popStringFilter(self, field):
        eqVals = self.popFilter(field, 'eq')
        if eqVals and len(eqVals) == 1:
            return eqVals[0]
        return None

    def popIntegerFilter(self, field):
        eqVals = self.popFilter(field, 'eq')
        if eqVals and len(eqVals) == 1:
            try:
                return int(eqVals[0])
            except ValueError as e:
                raise ValueError(
                    f"Filter value for {field} should be integer, but got: {eqVals[0]}"
                ) from e
        return None

    def removePagination(self):
        self.limit = self.offset = None

    def removeOrder(self):
        self.order = None

    def popField(self, field):
        try:
            i = self.fields.index(field)
        except ValueError:
            return False
        del self.fields[i]
        return True

    def findColumn(self, query, field):
        # will throw key error if field not in mapping
        mapped = self.fieldMapping[field]
        for col in query.inner_columns:
            if str(col) == mapped:
                return col
        raise KeyError(f"unable to find field {field} in query")

    def applyFilterToSQLQuery(self, query, f):
        field = f.field
        col = self.findColumn(query, field)
        # as sqlalchemy is overriding python operators, we can just use the same
        # python code generated by the filter
        return query.where(f.getOperator(sqlMode=True)(col, f.values))

    def applyOrderToSQLQuery(self, query, o):
        reverse = False
        if o.startswith('-'):
            reverse = True
            o = o[1:]
        col = self.findColumn(query, o)
        if reverse:
            col = col.desc()
        return query.order_by(col)

    def applyToSQLQuery(self, query):
        filters = self.filters
        order = self.order
        unmatched_filters = []
        unmatched_order = []
        # apply the filters if the name of field is found in the model, and
        # db2data
        for f in filters:
            try:
                query = self.applyFilterToSQLQuery(query, f)
            except KeyError:
                # if filter is unmatched, we will do the filtering manually in
                # self.apply
                unmatched_filters.append(f)

        # apply order if necessary
        if order:
            for o in order:
                try:
                    query = self.applyOrderToSQLQuery(query, o)
                except KeyError:
                    # if order is unmatched, we will do the ordering manually
                    # in self.apply
                    unmatched_order.append(o)

        # we cannot limit in sql if there is missing filtering or ordering
        if unmatched_filters or unmatched_order:
            if self.offset is not None or self.limit is not None:
                log.msg(
                    "Warning: limited data api query is not backed by db "
                    "because of following filters",
                    unmatched_filters,
                    unmatched_order,
                )
            self.filters = unmatched_filters
            self.order = tuple(unmatched_order)
            return query, None
        count_query = sa.select(sa.func.count()).select_from(query.alias('query'))
        self.order = None
        self.filters = []
        # finally, slice out the limit/offset
        if self.offset is not None:
            query = query.offset(self.offset)
            self.offset = None

        if self.limit is not None:
            query = query.limit(self.limit)
            self.limit = None

        return query, count_query

    def thd_execute(self, conn, q, dictFromRow):
        offset = self.offset
        limit = self.limit
        q, qc = self.applyToSQLQuery(q)
        res = conn.execute(q)
        rv = [dictFromRow(row) for row in res.fetchall()]

        if qc is not None and (offset or limit):
            total = conn.execute(qc).scalar()
            rv = base.ListResult(rv)
            rv.offset = offset
            rv.total = total
            rv.limit = limit
        return rv

    def apply(self, data):
        if data is None:
            return data

        if self.fields:
            fields = set(self.fields)

            def includeFields(d):
                if isinstance(d, dict):
                    return dict((k, v) for k, v in d.items() if k in fields)
                elif dataclasses.is_dataclass(d):
                    raise TypeError("includeFields can't filter fields of dataclasses")
                raise NotSupportedFieldTypeError(d)

            applyFields = includeFields
        else:
            fields = None
            applyFields = None

        if isinstance(data, dict) or dataclasses.is_dataclass(data):
            # item details
            if fields:
                data = applyFields(data)
            return data
        else:
            filters = self.filters
            order = self.order

            # item collection
            if isinstance(data, base.ListResult):
                # if pagination was applied, then fields, etc. must be empty
                assert not fields and not order and not filters, (
                    "endpoint must apply fields, order, and filters if it performs pagination"
                )
                offset = data.offset
                total = data.total
                limit = data.limit
            else:
                offset = None
                total = None
                limit = None

            if fields:
                data = (applyFields(d) for d in data)

            # link the filters together and then flatten to list
            for f in self.filters:
                data = f.apply(data)
            data = list(data)

            if total is None:
                total = len(data)

            if self.order:

                def keyFunc(elem, order=self.order):
                    """
                    Do a multi-level sort by passing in the keys
                    to sort by.

                    @param elem: each item in the list to sort.
                    @param order: a list of keys to sort by, such as:
                                ('lastName', 'firstName', 'age')
                    @return: a key used by sorted(). This will be a
                             list such as:
                             [a['lastName', a['firstName'], a['age']]
                    @rtype: a C{list}
                    """
                    compareKey = []
                    for k in order:
                        doReverse = False
                        if k[0] == '-':
                            # If we get a key '-lastName',
                            # it means sort by 'lastName' in reverse.
                            k = k[1:]
                            doReverse = True
                        val = NoneComparator(_data_getter(elem, k))
                        if doReverse:
                            val = ReverseComparator(val)
                        compareKey.append(val)
                    return compareKey

                data.sort(key=keyFunc)

            # finally, slice out the limit/offset
            if self.offset is not None or self.limit is not None:
                if offset is not None or limit is not None:
                    raise AssertionError("endpoint must clear offset/limit")
                end = (self.offset or 0) + self.limit if self.limit is not None else None
                data = data[self.offset : end]
                offset = self.offset
                limit = self.limit

            rv = base.ListResult(data)
            rv.offset = offset
            rv.total = total
            rv.limit = limit
            return rv


# a resultSpec which does not implement filtering in python (for tests)
class OptimisedResultSpec(ResultSpec):
    def apply(self, data):
        return data



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.python.failure import Failure
from twisted.web.server import Site


class SiteWithClose(Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._protocols = []

    def buildProtocol(self, addr):
        p = super().buildProtocol(addr)
        self._protocols.append(p)
        return p

    def close_connections(self):
        for p in self._protocols:
            p.connectionLost(Failure(RuntimeError("Closing down at the end of test")))
            # There is currently no other way to force all pending server-side connections to
            # close.
            p._channel.transport.connectionLost(
                Failure(RuntimeError("Closing down at the end of test"))
            )
        self._protocols = []



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Support for changes in the database
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer
from twisted.python import deprecate
from twisted.python import log
from twisted.python import versions

from buildbot.db import base
from buildbot.util import datetime2epoch
from buildbot.util import epoch2datetime
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime
    from typing import Any
    from typing import Iterable
    from typing import Literal


@dataclass
class ChangeModel:
    changeid: int
    author: str
    committer: str | None
    comments: str
    branch: str | None
    revision: str | None
    revlink: str | None
    when_timestamp: datetime.datetime
    category: str | None
    sourcestampid: int
    parent_changeids: list[int] = field(default_factory=list)
    repository: str = ''
    codebase: str = ''
    project: str = ''

    files: list[str] = field(default_factory=list)
    properties: dict[str, tuple[Any, Literal["Change"]]] = field(default_factory=dict)

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'ChangesConnectorComponent '
                'getChange, getChangesForBuild, getChangeFromSSid, and getChanges '
                'no longer return Change as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), ChangeModel)
class ChDict(ChangeModel):
    pass


class ChangesConnectorComponent(base.DBConnectorComponent):
    def getParentChangeIds(
        self, branch: str | None, repository: str, project: str, codebase: str
    ) -> defer.Deferred[list[int]]:
        def thd(conn) -> list[int]:
            changes_tbl = self.db.model.changes
            q = (
                sa.select(
                    changes_tbl.c.changeid,
                )
                .where(
                    changes_tbl.c.branch == branch,
                    changes_tbl.c.repository == repository,
                    changes_tbl.c.project == project,
                    changes_tbl.c.codebase == codebase,
                )
                .order_by(
                    sa.desc(changes_tbl.c.changeid),
                )
                .limit(1)
            )
            parent_id = conn.scalar(q)
            return [parent_id] if parent_id else []

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def addChange(
        self,
        author: str | None = None,
        committer: str | None = None,
        files: list[str] | None = None,
        comments: str | None = None,
        is_dir: None = None,
        revision: str | None = None,
        when_timestamp: datetime.datetime | None = None,
        branch: str | None = None,
        category: str | None = None,
        revlink: str | None = '',
        properties: dict[str, tuple[Any, Literal['Change']]] | None = None,
        repository: str = '',
        codebase: str = '',
        project: str = '',
        uid: int | None = None,
        _test_changeid: int | None = None,
    ):
        assert project is not None, "project must be a string, not None"
        assert repository is not None, "repository must be a string, not None"

        if is_dir is not None:
            log.msg("WARNING: change source is providing deprecated value is_dir (ignored)")
        if when_timestamp is None:
            when_timestamp = epoch2datetime(self.master.reactor.seconds())

        if author is None:
            author = ''
        if comments is None:
            comments = ''
        if properties is None:
            properties = {}

        # verify that source is 'Change' for each property
        for pv in properties.values():
            assert pv[1] == 'Change', "properties must be qualified with source 'Change'"

        ch_tbl = self.db.model.changes

        self.checkLength(ch_tbl.c.author, author)
        self.checkLength(ch_tbl.c.committer, committer)
        self.checkLength(ch_tbl.c.branch, branch)
        self.checkLength(ch_tbl.c.revision, revision)
        self.checkLength(ch_tbl.c.revlink, revlink)
        self.checkLength(ch_tbl.c.category, category)
        self.checkLength(ch_tbl.c.repository, repository)
        self.checkLength(ch_tbl.c.project, project)

        # calculate the sourcestamp first, before adding it
        ssid = yield self.db.sourcestamps.findSourceStampId(
            revision=revision,
            branch=branch,
            repository=repository,
            codebase=codebase,
            project=project,
        )

        parent_changeids = yield self.getParentChangeIds(branch, repository, project, codebase)
        # Someday, changes will have multiple parents.
        # But for the moment, a Change can only have 1 parent
        parent_changeid = parent_changeids[0] if parent_changeids else None

        def thd(conn) -> int:
            # note that in a read-uncommitted database like SQLite this
            # transaction does not buy atomicity - other database users may
            # still come across a change without its files, properties,
            # etc.  That's OK, since we don't announce the change until it's
            # all in the database, but beware.

            transaction = conn.begin()

            insert_value = {
                "author": author,
                "committer": committer,
                "comments": comments,
                "branch": branch,
                "revision": revision,
                "revlink": revlink,
                "when_timestamp": datetime2epoch(when_timestamp),
                "category": category,
                "repository": repository,
                "codebase": codebase,
                "project": project,
                "sourcestampid": ssid,
                "parent_changeids": parent_changeid,
            }

            if _test_changeid is not None:
                insert_value['changeid'] = _test_changeid

            r = conn.execute(ch_tbl.insert(), [insert_value])
            changeid = r.inserted_primary_key[0]
            if files:
                tbl = self.db.model.change_files
                for f in files:
                    self.checkLength(tbl.c.filename, f)
                conn.execute(tbl.insert(), [{"changeid": changeid, "filename": f} for f in files])
            if properties:
                tbl = self.db.model.change_properties
                inserts = [
                    {"changeid": changeid, "property_name": k, "property_value": json.dumps(v)}
                    for k, v in properties.items()
                ]
                for i in inserts:
                    self.checkLength(tbl.c.property_name, i['property_name'])

                conn.execute(tbl.insert(), inserts)
            if uid:
                ins = self.db.model.change_users.insert()
                conn.execute(ins, {"changeid": changeid, "uid": uid})

            transaction.commit()

            return changeid

        return (yield self.db.pool.do(thd))

    @base.cached("chdicts")
    def getChange(self, changeid: int) -> defer.Deferred[ChangeModel | None]:
        assert changeid >= 0

        def thd(conn) -> ChangeModel | None:
            # get the row from the 'changes' table
            changes_tbl = self.db.model.changes
            q = changes_tbl.select().where(changes_tbl.c.changeid == changeid)
            rp = conn.execute(q)
            row = rp.fetchone()
            if not row:
                return None
            # and fetch the ancillary data (files, properties)
            return self._thd_model_from_row(conn, row)

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getChangesForBuild(self, buildid: int):
        assert buildid > 0

        gssfb = self.master.db.sourcestamps.getSourceStampsForBuild
        changes: list[ChangeModel] = []
        currentBuild = yield self.master.db.builds.getBuild(buildid)
        fromChanges: dict[str, ChangeModel | None] = {}
        toChanges: dict[str, ChangeModel] = {}
        ssBuild = yield gssfb(buildid)
        for ss in ssBuild:
            fromChanges[ss.codebase] = yield self.getChangeFromSSid(ss.ssid)

        # Get the last successful build on the same builder
        previousBuild = yield self.master.db.builds.getPrevSuccessfulBuild(
            currentBuild.builderid, currentBuild.number, ssBuild
        )
        if previousBuild:
            for ss in (yield gssfb(previousBuild.id)):
                ss_change = yield self.getChangeFromSSid(ss.ssid)
                if ss_change:
                    toChanges[ss.codebase] = ss_change

        # For each codebase, append changes until we match the parent
        for cb, change in fromChanges.items():
            if not change:
                continue

            to_cb_change = toChanges.get(cb)
            to_cb_changeid = to_cb_change.changeid if to_cb_change is not None else None
            if to_cb_changeid is not None and to_cb_changeid == change.changeid:
                continue

            changes.append(change)
            while change.parent_changeids and to_cb_changeid not in change.parent_changeids:
                # For the moment, a Change only have 1 parent.
                change = yield self.master.db.changes.getChange(change.parent_changeids[0])
                # http://trac.buildbot.net/ticket/3461 sometimes,
                # parent_changeids could be corrupted
                if change is None:
                    break
                changes.append(change)

        return changes

    def getChangeFromSSid(self, sourcestampid: int) -> defer.Deferred[ChangeModel | None]:
        assert sourcestampid >= 0

        def thd(conn) -> ChangeModel | None:
            # get the row from the 'changes' table
            changes_tbl = self.db.model.changes
            q = changes_tbl.select().where(changes_tbl.c.sourcestampid == sourcestampid)
            # if there are multiple changes for this ssid, get the most recent one
            q = q.order_by(changes_tbl.c.changeid.desc())
            q = q.limit(1)
            rp = conn.execute(q)
            row = rp.fetchone()
            if not row:
                return None
            # and fetch the ancillary data (files, properties)
            return self._thd_model_from_row(conn, row)

        return self.db.pool.do(thd)

    def getChangeUids(self, changeid: int) -> defer.Deferred[list[int]]:
        assert changeid >= 0

        def thd(conn) -> list[int]:
            cu_tbl = self.db.model.change_users
            q = cu_tbl.select().where(cu_tbl.c.changeid == changeid)
            res = conn.execute(q)
            rows = res.fetchall()
            row_uids = [row.uid for row in rows]
            return row_uids

        return self.db.pool.do(thd)

    def _getDataFromRow(self, row):
        return row.changeid

    @defer.inlineCallbacks
    def getChanges(self, resultSpec=None):
        def thd(conn) -> Iterable[int]:
            # get the changeids from the 'changes' table
            changes_tbl = self.db.model.changes

            if resultSpec is not None:
                q = changes_tbl.select()
                return reversed(resultSpec.thd_execute(conn, q, self._getDataFromRow))

            q = sa.select(changes_tbl.c.changeid)
            rp = conn.execute(q)
            changeids = [self._getDataFromRow(row) for row in rp]
            rp.close()
            return list(changeids)

        changeids = yield self.db.pool.do(thd)

        changes = yield defer.gatherResults(
            [self.getChange(changeid) for changeid in changeids], consumeErrors=True
        )

        return changes

    def getChangesCount(self) -> defer.Deferred[int]:
        def thd(conn) -> int:
            changes_tbl = self.db.model.changes
            q = sa.select(sa.func.count()).select_from(changes_tbl)
            rp = conn.execute(q)
            r = 0
            for row in rp:
                r = row[0]
            rp.close()
            return int(r)

        return self.db.pool.do(thd)

    def getLatestChangeid(self) -> defer.Deferred[int | None]:
        def thd(conn) -> int:
            changes_tbl = self.db.model.changes
            q = (
                sa.select(
                    changes_tbl.c.changeid,
                )
                .order_by(
                    sa.desc(changes_tbl.c.changeid),
                )
                .limit(1)
            )
            return conn.scalar(q)

        return self.db.pool.do(thd)

    # utility methods

    @defer.inlineCallbacks
    def pruneChanges(self, changeHorizon: int):
        """
        Called periodically by DBConnector, this method deletes changes older
        than C{changeHorizon}.
        """

        if not changeHorizon:
            return

        def thd(conn) -> None:
            changes_tbl = self.db.model.changes

            # First, get the list of changes to delete.  This could be written
            # as a subquery but then that subquery would be run for every
            # table, which is very inefficient; also, MySQL's subquery support
            # leaves much to be desired, and doesn't support this particular
            # form.
            q = (
                sa.select(
                    changes_tbl.c.changeid,
                )
                .order_by(
                    sa.desc(changes_tbl.c.changeid),
                )
                .offset(changeHorizon)
            )
            res = conn.execute(q)
            ids_to_delete = [r.changeid for r in res]

            # and delete from all relevant tables, in dependency order
            for table_name in (
                'scheduler_changes',
                'change_files',
                'change_properties',
                'changes',
                'change_users',
            ):
                remaining = ids_to_delete[:]
                while remaining:
                    batch = remaining[:100]
                    remaining = remaining[100:]
                    table = self.db.model.metadata.tables[table_name]
                    conn.execute(table.delete().where(table.c.changeid.in_(batch)))

        yield self.db.pool.do_with_transaction(thd)

    def _thd_model_from_row(self, conn, ch_row) -> ChangeModel:
        # This method must be run in a db.pool thread
        change_files_tbl = self.db.model.change_files
        change_properties_tbl = self.db.model.change_properties

        if ch_row.parent_changeids:
            parent_changeids = [ch_row.parent_changeids]
        else:
            parent_changeids = []

        chdict = ChangeModel(
            changeid=ch_row.changeid,
            parent_changeids=parent_changeids,
            author=ch_row.author,
            committer=ch_row.committer,
            comments=ch_row.comments,
            revision=ch_row.revision,
            when_timestamp=epoch2datetime(ch_row.when_timestamp),
            branch=ch_row.branch,
            category=ch_row.category,
            revlink=ch_row.revlink,
            repository=ch_row.repository,
            codebase=ch_row.codebase,
            project=ch_row.project,
            sourcestampid=int(ch_row.sourcestampid),
        )

        query = change_files_tbl.select().where(change_files_tbl.c.changeid == ch_row.changeid)
        rows = conn.execute(query)
        chdict.files.extend(r.filename for r in rows)

        # and properties must be given without a source, so strip that, but
        # be flexible in case users have used a development version where the
        # change properties were recorded incorrectly
        def split_vs(vs) -> tuple[Any, Literal["Change"]]:
            try:
                v, s = vs
                if s != "Change":
                    v, s = vs, "Change"
            except (ValueError, TypeError):
                v, s = vs, "Change"
            return v, s

        query = change_properties_tbl.select().where(
            change_properties_tbl.c.changeid == ch_row.changeid
        )
        rows = conn.execute(query)
        for r in rows:
            try:
                v, s = split_vs(json.loads(r.property_value))
                chdict.properties[r.property_name] = (v, s)
            except ValueError:
                pass

        return chdict



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import inspect
import sys
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Sequence
from typing import cast

from twisted.internet import defer
from twisted.internet import error
from twisted.python import deprecate
from twisted.python import log
from twisted.python import versions
from twisted.python.failure import Failure
from twisted.python.reflect import accumulateClassList
from twisted.web.util import formatFailure
from zope.interface import implementer

from buildbot import config
from buildbot import interfaces
from buildbot import util
from buildbot.config.checks import check_param_bool
from buildbot.config.checks import check_param_length
from buildbot.config.checks import check_param_number_none
from buildbot.config.checks import check_param_str
from buildbot.config.checks import check_param_str_none
from buildbot.db import model_config
from buildbot.interfaces import IRenderable
from buildbot.interfaces import WorkerSetupError
from buildbot.process import log as plog
from buildbot.process import properties
from buildbot.process import remotecommand
from buildbot.process import results
from buildbot.process.locks import get_real_locks_from_accesses

# (WithProperties used to be available in this module)
from buildbot.process.properties import WithProperties  # noqa: F401
from buildbot.process.results import ALL_RESULTS
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SKIPPED
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.process.results import statusToString
from buildbot.util import bytes2unicode
from buildbot.util import debounce
from buildbot.util import deferwaiter
from buildbot.util import flatten
from buildbot.util.test_result_submitter import TestResultSubmitter

if TYPE_CHECKING:
    from typing import TypeVar

    from twisted.internet.base import ReactorBase
    from typing_extensions import Self

    from buildbot.interfaces import IBuildStep
    from buildbot.locks import BaseLock
    from buildbot.master import BuildMaster
    from buildbot.process.build import Build
    from buildbot.process.log import StreamLog
    from buildbot.util.twisted import InlineCallbacksType
    from buildbot.worker.base import AbstractWorker
    from buildbot.worker.protocols.base import Connection

    BuildStepType = TypeVar('BuildStepType', bound="BuildStep")


class BuildStepFailed(Exception):
    pass


class BuildStepCancelled(Exception):
    # used internally for signalling
    pass


class CallableAttributeError(Exception):
    # attribute error raised from a callable run inside a property
    pass


@implementer(interfaces.IBuildStepFactory)
class _BuildStepFactory(util.ComparableMixin):
    """
    This is a wrapper to record the arguments passed to as BuildStep subclass.
    We use an instance of this class, rather than a closure mostly to make it
    easier to test that the right factories are getting created.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('factory', 'args', 'kwargs')

    def __init__(self, step_class: type[BuildStep], *args: Any, **kwargs: Any) -> None:
        self.step_class = step_class
        self.args = args
        self.kwargs = kwargs

    def buildStep(self) -> BuildStep:
        try:
            step = object.__new__(self.step_class)
            step._factory = self
            step.__init__(*self.args, **self.kwargs)  # type: ignore[misc]
            return step
        except Exception:
            log.msg(
                f"error while creating step, step_class={self.step_class}, args={self.args}, "
                f"kwargs={self.kwargs}"
            )
            raise


class BuildStepStatus:
    # used only for old-style steps
    pass


def get_factory_from_step_or_factory(
    step_or_factory: BuildStep | interfaces.IBuildStepFactory,
) -> interfaces.IBuildStepFactory:
    if hasattr(step_or_factory, 'get_step_factory'):
        factory = step_or_factory.get_step_factory()
    else:
        factory = step_or_factory
    # make sure the returned value actually implements IBuildStepFactory
    return interfaces.IBuildStepFactory(factory)


def create_step_from_step_or_factory(
    step_or_factory: BuildStep | interfaces.IBuildStepFactory,
) -> IBuildStep:
    return get_factory_from_step_or_factory(step_or_factory).buildStep()


class BuildStepWrapperMixin:
    __init_completed: bool = False

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__init_completed = True

    def __setattr__(self, name: str, value: Any) -> None:
        if self.__init_completed:
            config.error(
                "Changes to attributes of a BuildStep instance are ignored, this is a bug. "
                "Use set_step_arg(name, value) for that."
            )
        super().__setattr__(name, value)


# This is also needed for comparisons to work because ComparableMixin requires type(x) and
# x.__class__ to be equal in order to perform comparison at all.
_buildstep_wrapper_cache: dict[int, type] = {}


def _create_buildstep_wrapper_class(klass: type[BuildStepType]) -> type[BuildStepType]:
    class_id = id(klass)
    cached = _buildstep_wrapper_cache.get(class_id, None)
    if cached is not None:
        return cached

    wrapper = type(klass.__qualname__, (BuildStepWrapperMixin, klass), {})
    _buildstep_wrapper_cache[class_id] = wrapper
    return cast("type[BuildStepType]", wrapper)


@implementer(interfaces.IBuildStep)
class BuildStep(
    results.ResultComputingConfigMixin, properties.PropertiesMixin, util.ComparableMixin
):
    # Note that the BuildStep is at the same time a template from which per-build steps are
    # constructed. This works by creating a new IBuildStepFactory in __new__, retrieving it via
    # get_step_factory() and then calling buildStep() on that factory.

    alwaysRun: bool = False
    doStepIf: bool | Callable[[BuildStep], bool | defer.Deferred[bool]] = True
    hideStepIf: bool | Callable[[int, BuildStep], bool] = False
    compare_attrs: ClassVar[Sequence[str]] = ("_factory",)
    # properties set on a build step are, by nature, always runtime properties
    set_runtime_properties: bool = True
    _factory: _BuildStepFactory

    renderables: Sequence[str] = [
        *results.ResultComputingConfigMixin.resultConfig,
        'alwaysRun',
        'description',
        'descriptionDone',
        'descriptionSuffix',
        'doStepIf',
        'hideStepIf',
        'workdir',
    ]

    # '_params_names' holds a list of all the parameters we care about, to allow
    # users to instantiate a subclass of BuildStep with a mixture of
    # arguments, some of which are for us, some of which are for the subclass
    # (or a delegate of the subclass, like how ShellCommand delivers many
    # arguments to the RemoteShellCommand that it creates). Such delegating
    # subclasses will use this list to figure out which arguments are meant
    # for us and which should be given to someone else.
    _params_config: list[tuple[str, Callable | None]] = [
        ('alwaysRun', check_param_bool),
        ('description', None),
        ('descriptionDone', None),
        ('descriptionSuffix', None),
        ('doStepIf', None),
        ('flunkOnFailure', check_param_bool),
        ('flunkOnWarnings', check_param_bool),
        ('haltOnFailure', check_param_bool),
        ('updateBuildSummaryPolicy', None),
        ('hideStepIf', None),
        ('locks', None),
        ('logEncoding', None),
        ('name', check_param_str),
        ('progressMetrics', None),
        ('useProgress', None),
        ('warnOnFailure', check_param_bool),
        ('warnOnWarnings', check_param_bool),
        ('workdir', check_param_str_none),
    ]

    _params_names: list[str] = [arg for arg, _ in _params_config]

    name: str | IRenderable = "generic"
    description: str | list[str] | None = None  # set this to a list of short strings to override
    descriptionDone: str | list[str] | None = (
        None  # alternate description when the step is complete
    )
    descriptionSuffix: str | list[str] | None = None  # extra information to append to suffix
    updateBuildSummaryPolicy: list[int] | None | bool = None
    locks: list[str] | None = None
    _locks_to_acquire: list[tuple[BaseLock, Any]] = []
    progressMetrics: tuple[str, ...] = ()  # 'time' is implicit
    useProgress: bool = True  # set to False if step is really unpredictable
    build: Build | None = None
    step_status: None = None
    progress: None = None
    logEncoding: str | None = None
    cmd: remotecommand.RemoteCommand | None = None
    rendered: bool = False  # true if attributes are rendered
    _workdir: str | None = None
    _waitingForLocks: bool = False

    def __init__(self, **kwargs: Any) -> None:
        self.worker = None

        for p, check in self.__class__._params_config:
            if p in kwargs:
                value = kwargs.pop(p)
                if check is not None and not IRenderable.providedBy(value):
                    check(value, self.__class__, p)
                setattr(self, p, value)

        if kwargs:
            config.error(
                f"{self.__class__}.__init__ got unexpected keyword argument(s) {list(kwargs)}"
            )
        self._pendingLogObservers: list[tuple[str, interfaces.ILogObserver]] = []

        check_param_length(
            self.name, f'Step {self.__class__.__name__} name', model_config.step_name_length
        )

        if isinstance(self.description, str):
            self.description = [self.description]
        if isinstance(self.descriptionDone, str):
            self.descriptionDone = [self.descriptionDone]
        if isinstance(self.descriptionSuffix, str):
            self.descriptionSuffix = [self.descriptionSuffix]

        if self.updateBuildSummaryPolicy is None:
            # compute default value for updateBuildSummaryPolicy
            self.updateBuildSummaryPolicy = [EXCEPTION, RETRY, CANCELLED]
            if self.flunkOnFailure or self.haltOnFailure or self.warnOnFailure:
                self.updateBuildSummaryPolicy.append(FAILURE)
            if self.warnOnWarnings or self.flunkOnWarnings:
                self.updateBuildSummaryPolicy.append(WARNINGS)
        if self.updateBuildSummaryPolicy is False:
            self.updateBuildSummaryPolicy = []
        if self.updateBuildSummaryPolicy is True:
            self.updateBuildSummaryPolicy = ALL_RESULTS
        if not isinstance(self.updateBuildSummaryPolicy, list):
            config.error(
                "BuildStep updateBuildSummaryPolicy must be "
                "a list of result ids or boolean but it is "
                f"{self.updateBuildSummaryPolicy!r}"
            )
        self._acquiringLocks: list[tuple[BaseLock, str, defer.Deferred]] = []
        self.stopped = False
        self.timed_out = False
        self.max_lines_reached = False
        self.master: None | BuildMaster = None
        self.statistics: dict[str, int] = {}
        self.logs: dict[str, plog.Log] = {}
        self._running = False
        self.stepid = None
        self.results: int | None = None
        self._start_unhandled_deferreds = None
        self._interrupt_deferwaiter = deferwaiter.DeferWaiter()

        def get_master_reactor() -> ReactorBase:
            assert self.master is not None
            return self.master.reactor

        self._update_summary_debouncer = debounce.Debouncer(
            1.0, self._update_summary_impl, get_master_reactor, until_idle=False
        )
        self._test_result_submitters: dict[int, TestResultSubmitter] = {}

    def __new__(klass: type[Self], *args: Any, **kwargs: Any) -> Self:
        # The following code prevents changing BuildStep attributes after an instance
        # is created during config time. Such attribute changes don't affect the factory,
        # so they will be lost when actual build step is created.
        #
        # This is implemented by dynamically creating a subclass that disallows attribute
        # writes after __init__ completes.
        self = object.__new__(_create_buildstep_wrapper_class(klass))
        self._factory = _BuildStepFactory(klass, *args, **kwargs)
        return self

    def is_exact_step_class(self, klass: type[BuildStep]) -> bool:
        # Due to wrapping BuildStep in __new__, it's not possible to compare self.__class__ to
        # check if self is an instance of some class (but not subclass).
        if self.__class__ is klass:
            return True
        mro = self.__class__.mro()
        if len(mro) >= 3 and mro[1] is BuildStepWrapperMixin and mro[2] is klass:
            return True
        return False

    def __str__(self) -> str:
        args = [repr(x) for x in self._factory.args]
        args.extend([str(k) + "=" + repr(v) for k, v in self._factory.kwargs.items()])
        return f'{self.__class__.__name__}({", ".join(args)})'

    __repr__ = __str__

    def setBuild(self, build: Build) -> None:
        self.build = build
        self.master = self.build.master

    def setWorker(self, worker: AbstractWorker) -> None:
        self.worker = worker

    @deprecate.deprecated(versions.Version("buildbot", 0, 9, 0))
    def setDefaultWorkdir(self, workdir: str) -> None:
        if self._workdir is None:
            self._workdir = workdir

    @property
    def workdir(self) -> str | None:
        # default the workdir appropriately
        if self._workdir is not None or self.build is None:
            return self._workdir
        else:
            # see :ref:`Factory-Workdir-Functions` for details on how to
            # customize this
            if callable(self.build.workdir):
                try:
                    return self.build.workdir(self.build.sources)
                except AttributeError as e:
                    # if the callable raises an AttributeError
                    # python thinks it is actually workdir that is not existing.
                    # python will then swallow the attribute error and call
                    # __getattr__ from worker_transition
                    _, _, traceback = sys.exc_info()
                    raise CallableAttributeError(e).with_traceback(traceback) from e
                    # we re-raise the original exception by changing its type,
                    # but keeping its stacktrace
            else:
                return self.build.workdir

    @workdir.setter
    def workdir(self, workdir: str) -> None:
        self._workdir = workdir

    def getProperties(self) -> properties.Properties:
        assert self.build is not None
        return self.build.getProperties()

    def get_step_factory(self) -> _BuildStepFactory:
        return self._factory

    def set_step_arg(self, name: str, value: Any) -> None:
        self._factory.kwargs[name] = value
        # check if buildstep can still be constructed with the new arguments
        try:
            self._factory.buildStep()
        except Exception:
            log.msg(f"Cannot set step factory attribute {name} to {value}: step creation fails")
            raise

    def setupProgress(self) -> None:
        # this function temporarily does nothing
        pass

    def setProgress(self, metric: str, value: int) -> None:
        # this function temporarily does nothing
        pass

    def getCurrentSummary(self) -> dict[str, str]:
        if self.description is not None:
            stepsumm = util.join_list(self.description)
            if self.descriptionSuffix:
                stepsumm += ' ' + util.join_list(self.descriptionSuffix)
        else:
            stepsumm = 'running'
        return {'step': stepsumm}

    def getResultSummary(self) -> dict[str, str]:
        if self.descriptionDone is not None or self.description is not None:
            stepsumm = util.join_list(self.descriptionDone or self.description)
            if self.descriptionSuffix:
                stepsumm += ' ' + util.join_list(self.descriptionSuffix)
        else:
            stepsumm = 'finished'

        if self.results != SUCCESS:
            stepsumm += f' ({statusToString(self.results)})'
            if self.timed_out:
                stepsumm += " (timed out)"
            elif self.max_lines_reached:
                stepsumm += " (max lines reached)"

        if self.build is not None:
            stepsumm = self.build.properties.cleanupTextFromSecrets(stepsumm)
        return {'step': stepsumm}

    @defer.inlineCallbacks
    def getBuildResultSummary(self) -> InlineCallbacksType[dict[str, str]]:
        summary: dict[str, str] = yield self.getResultSummary()  # type: ignore[misc]
        assert isinstance(self.updateBuildSummaryPolicy, list)
        if (
            self.results in self.updateBuildSummaryPolicy
            and 'build' not in summary
            and 'step' in summary
        ):
            summary['build'] = summary['step']
        return summary

    def updateSummary(self) -> None:
        self._update_summary_debouncer()

    @defer.inlineCallbacks
    def _update_summary_impl(self) -> InlineCallbacksType[None]:
        def methodInfo(m: Callable) -> str:
            lines = inspect.getsourcelines(m)
            return "\nat {}:{}:\n {}".format(
                inspect.getsourcefile(m), lines[1], "\n".join(lines[0])
            )

        if not self._running:
            summary = yield self.getResultSummary()  # type: ignore[misc]
            if not isinstance(summary, dict):
                raise TypeError(
                    'getResultSummary must return a dictionary: '
                    + methodInfo(self.getResultSummary)
                )
        else:
            summary = yield self.getCurrentSummary()  # type: ignore[misc]
            if not isinstance(summary, dict):
                raise TypeError(
                    'getCurrentSummary must return a dictionary: '
                    + methodInfo(self.getCurrentSummary)
                )

        stepResult = summary.get('step', 'finished')
        if not isinstance(stepResult, str):
            raise TypeError(f"step result string must be unicode (got {stepResult!r})")
        if self.stepid is not None:
            assert self.build is not None
            stepResult = self.build.properties.cleanupTextFromSecrets(stepResult)
            yield self.master.data.updates.setStepStateString(self.stepid, stepResult)

        if not self._running:
            buildResult = summary.get('build', None)
            if buildResult and not isinstance(buildResult, str):
                raise TypeError("build result string must be unicode")

    @defer.inlineCallbacks
    def addStep(self) -> InlineCallbacksType[None]:
        # create and start the step, noting that the name may be altered to
        # ensure uniqueness
        assert self.build is not None
        name = yield cast(defer.Deferred[str], self.build.render(self.name))
        name = self.build.setUniqueStepName(name)
        self.name = name
        assert self.master is not None
        self.stepid, self.number, self.name = yield self.master.data.updates.addStep(
            buildid=self.build.buildid, name=util.bytes2unicode(name)
        )

    @defer.inlineCallbacks
    def startStep(self, remote: Connection) -> InlineCallbacksType[int]:
        self.remote = remote

        yield self.addStep()
        assert self.master is not None
        started_at = int(self.master.reactor.seconds())
        yield self.master.data.updates.startStep(self.stepid, started_at=started_at)

        try:
            yield self._render_renderables()
            # we describe ourselves only when renderables are interpolated
            self.updateSummary()

            # check doStepIf (after rendering)
            if isinstance(self.doStepIf, bool):
                doStep = self.doStepIf
            else:
                doStep = cast(bool, (yield self.doStepIf(self)))

            if doStep:
                yield self._setup_locks()

                # set up locks
                if self._locks_to_acquire:
                    yield self.acquireLocks()

                    if self.stopped:
                        raise BuildStepCancelled

                    locks_acquired_at = int(self.master.reactor.seconds())
                    assert self.build is not None
                    yield defer.DeferredList(
                        [
                            self.master.data.updates.set_step_locks_acquired_at(
                                self.stepid, locks_acquired_at=locks_acquired_at
                            ),
                            self.master.data.updates.add_build_locks_duration(
                                self.build.buildid, duration_s=locks_acquired_at - started_at
                            ),
                        ],
                        consumeErrors=True,
                    )
                else:
                    yield self.master.data.updates.set_step_locks_acquired_at(
                        self.stepid, locks_acquired_at=started_at
                    )

                    if self.stopped:
                        raise BuildStepCancelled

                yield self.addTestResultSets()
                try:
                    self._running = True
                    self.results = yield self.run()
                finally:
                    self._running = False
            else:
                self.results = SKIPPED

        # NOTE: all of these `except` blocks must set self.results immediately!
        except BuildStepCancelled:
            self.results = CANCELLED

        except BuildStepFailed:
            self.results = FAILURE

        except error.ConnectionLost:
            self.results = RETRY

        except Exception:
            self.results = EXCEPTION
            why = Failure()
            log.err(why, "BuildStep.failed; traceback follows")
            yield self.addLogWithFailure(why)

        if self.stopped and self.results != RETRY:
            # We handle this specially because we don't care about
            # the return code of an interrupted command; we know
            # that this should just be exception due to interrupt
            # At the same time we must respect RETRY status because it's used
            # to retry interrupted build due to some other issues for example
            # due to worker lost
            if self.results != CANCELLED:
                self.results = EXCEPTION

        # determine whether we should hide this step
        hidden = self.hideStepIf
        if callable(self.hideStepIf):
            try:
                assert self.results is not None
                hidden = self.hideStepIf(self.results, self)
            except Exception:
                why = Failure()
                log.err(why, "hidden callback failed; traceback follows")
                yield self.addLogWithFailure(why)
                self.results = EXCEPTION
                hidden = False

        # perform final clean ups
        success = yield self._cleanup_logs()
        if not success:
            self.results = EXCEPTION

        # update the summary one last time, make sure that completes,
        # and then don't update it any more.
        self.updateSummary()
        yield self._update_summary_debouncer.stop()

        for sub in self._test_result_submitters.values():
            yield sub.finish()

        self.releaseLocks()

        yield self.master.data.updates.finishStep(self.stepid, self.results, hidden)

        assert self.results is not None

        return self.results

    @defer.inlineCallbacks
    def _setup_locks(self) -> InlineCallbacksType[None]:
        self._locks_to_acquire = yield get_real_locks_from_accesses(self.locks, self.build)

        assert self.build is not None
        if self.build._locks_to_acquire:
            build_locks = [l for l, _ in self.build._locks_to_acquire]
            for l, _ in self._locks_to_acquire:
                if l in build_locks:
                    log.err(
                        f"{self}: lock {l} is claimed by both a Step ({self}) and the"
                        f" parent Build ({self.build})"
                    )
                    raise RuntimeError(f"lock claimed by both Step and Build ({l})")

    @defer.inlineCallbacks
    def _render_renderables(self) -> InlineCallbacksType[None]:
        # render renderables in parallel
        renderables: list[str] = []
        accumulateClassList(self.__class__, 'renderables', renderables)

        def setRenderable(res: Any, attr: str) -> None:
            setattr(self, attr, res)

        dl = []
        for renderable in renderables:
            assert self.build is not None
            d = self.build.render(getattr(self, renderable))
            d.addCallback(setRenderable, renderable)
            dl.append(d)
        yield defer.gatherResults(dl, consumeErrors=True)
        self.rendered = True

    def setBuildData(self, name: str, value: bytes, source: str) -> defer.Deferred:
        # returns a Deferred that yields nothing
        assert self.master is not None
        assert self.build is not None
        return self.master.data.updates.setBuildData(self.build.buildid, name, value, source)

    @defer.inlineCallbacks
    def _cleanup_logs(self) -> InlineCallbacksType[bool]:
        # Wait until any in-progress interrupt() to finish (that function may add new logs)
        yield self._interrupt_deferwaiter.wait()

        all_success = True
        not_finished_logs = [v for (k, v) in self.logs.items() if not v.finished]
        finish_logs = yield defer.DeferredList(
            [v.finish() for v in not_finished_logs], consumeErrors=True
        )
        for success, res in finish_logs:
            if not success:
                log.err(res, "when trying to finish a log")
                all_success = False

        for log_ in self.logs.values():
            if log_.had_errors():
                all_success = False

        return all_success

    def addTestResultSets(self) -> defer.Deferred:
        return defer.succeed(None)

    @defer.inlineCallbacks
    def addTestResultSet(
        self, description: str, category: str, value_unit: str
    ) -> InlineCallbacksType[int]:
        sub = TestResultSubmitter()
        yield sub.setup(self, description, category, value_unit)
        setid = sub.get_test_result_set_id()
        self._test_result_submitters[setid] = sub
        return setid

    def addTestResult(
        self,
        setid: int,
        value: float,
        test_name: str | None = None,
        test_code_path: str | None = None,
        line: int | None = None,
        duration_ns: int | None = None,
    ) -> None:
        self._test_result_submitters[setid].add_test_result(
            value,
            test_name=test_name,
            test_code_path=test_code_path,
            line=line,
            duration_ns=duration_ns,
        )

    def acquireLocks(self, res: Any = None) -> defer.Deferred[None | BaseLock]:
        if not self._locks_to_acquire:
            return defer.succeed(None)
        if self.stopped:
            return defer.succeed(None)
        log.msg(f"acquireLocks(step {self}, locks {self._locks_to_acquire})")
        for lock, access in self._locks_to_acquire:
            for waited_lock, _, _ in self._acquiringLocks:
                if lock is waited_lock:
                    continue

            if not lock.isAvailable(self, access):
                self._waitingForLocks = True
                log.msg(f"step {self} waiting for lock {lock}")
                d = lock.waitUntilMaybeAvailable(self, access)
                self._acquiringLocks.append((lock, access, d))
                d.addCallback(self.acquireLocks)
                return d
        # all locks are available, claim them all
        for lock, access in self._locks_to_acquire:
            lock.claim(self, access)
        self._acquiringLocks = []
        self._waitingForLocks = False
        return defer.succeed(None)

    def run(self) -> defer.Deferred[int]:
        raise NotImplementedError("A custom build step must implement run()")

    @defer.inlineCallbacks
    def _maybe_interrupt_cmd(self, reason: str | Failure) -> InlineCallbacksType[None]:
        if not self.cmd:
            return

        try:
            yield self.cmd.interrupt(reason)
        except Exception as e:
            log.err(e, 'while cancelling command')

    def interrupt(self, reason: str | Failure) -> defer.Deferred[None]:
        # Note that this method may be run outside usual step lifecycle (e.g. after run() has
        # already completed), so extra care needs to be taken to prevent race conditions.
        return self._interrupt_deferwaiter.add(self._interrupt_impl(reason))

    @defer.inlineCallbacks
    def _interrupt_impl(self, reason: str | Failure) -> InlineCallbacksType[None]:
        if self.stopped:
            # If we are in the process of interruption and connection is lost then we must tell
            # the command not to wait for the interruption to complete.
            if isinstance(reason, Failure) and reason.check(error.ConnectionLost):
                yield self._maybe_interrupt_cmd(reason)
            return

        self.stopped = True
        if self._acquiringLocks:
            for lock, access, d in self._acquiringLocks:
                lock.stopWaitingUntilAvailable(self, access, d)
            self._acquiringLocks = []

        log_name = "cancelled while waiting for locks" if self._waitingForLocks else "cancelled"
        yield self.addCompleteLog(log_name, str(reason))
        yield self._maybe_interrupt_cmd(reason)

    def releaseLocks(self) -> None:
        log.msg(f"releaseLocks({self}): {self._locks_to_acquire}")
        for lock, access in self._locks_to_acquire:
            if lock.isOwner(self, access):
                lock.release(self, access)
            else:
                # This should only happen if we've been interrupted
                assert self.stopped

    # utility methods that BuildSteps may find useful

    def workerVersion(self, command: str, oldversion: str | None = None) -> str | None:
        assert self.build is not None
        return self.build.getWorkerCommandVersion(command, oldversion)

    def workerVersionIsOlderThan(self, command: str, minversion: str) -> bool:
        assert self.build is not None
        sv = self.build.getWorkerCommandVersion(command, None)
        if sv is None:
            return True
        if [int(s) for s in sv.split(".")] < [int(m) for m in minversion.split(".")]:
            return True
        return False

    def checkWorkerHasCommand(self, command: str) -> None:
        if not self.workerVersion(command):
            message = f"worker is too old, does not know about {command}"
            raise WorkerSetupError(message)

    def getWorkerName(self) -> str | None:
        assert self.build is not None
        return self.build.getWorkerName()

    def addLog(
        self, name: str, type: str = 's', logEncoding: str | None = None
    ) -> defer.Deferred[plog.Log]:
        if self.stepid is None:
            raise BuildStepCancelled
        assert self.master is not None
        d = self.master.data.updates.addLog(self.stepid, util.bytes2unicode(name), str(type))

        @d.addCallback
        def newLog(logid: int) -> plog.Log:
            return self._newLog(name, type, logid, logEncoding)

        return d

    def getLog(self, name: str) -> plog.Log:
        return self.logs[name]

    @defer.inlineCallbacks
    def addCompleteLog(self, name: str, text: str | bytes) -> InlineCallbacksType[None]:
        if self.stepid is None:
            raise BuildStepCancelled
        assert self.master is not None
        logid = yield self.master.data.updates.addLog(self.stepid, util.bytes2unicode(name), 't')
        _log = cast(plog.PlainLog, self._newLog(name, 't', logid))
        yield _log.addContent(text)
        yield _log.finish()

    @defer.inlineCallbacks
    def addHTMLLog(self, name: str, html: str | bytes) -> InlineCallbacksType[None]:
        if self.stepid is None:
            raise BuildStepCancelled
        assert self.master is not None
        logid = yield self.master.data.updates.addLog(self.stepid, util.bytes2unicode(name), 'h')
        _log = cast(plog.HtmlLog, self._newLog(name, 'h', logid))
        yield _log.addContent(bytes2unicode(html))
        yield _log.finish()

    @defer.inlineCallbacks
    def addLogWithFailure(self, why: Failure, logprefix: str = "") -> InlineCallbacksType[None]:
        # helper for showing exceptions to the users
        try:
            yield self.addCompleteLog(logprefix + "err.text", why.getTraceback())
            yield self.addHTMLLog(logprefix + "err.html", formatFailure(why))
        except Exception:
            log.err(Failure(), "error while formatting exceptions")

    def addLogWithException(self, why: Failure, logprefix: str = "") -> defer.Deferred[None]:
        return self.addLogWithFailure(Failure(why), logprefix)

    def addLogObserver(self, logname: str, observer: interfaces.ILogObserver) -> None:
        assert interfaces.ILogObserver.providedBy(observer)
        observer.setStep(self)
        self._pendingLogObservers.append((logname, observer))
        self._connectPendingLogObservers()

    def _newLog(self, name: str, type: str, logid: int, logEncoding: str | None = None) -> plog.Log:
        if not logEncoding:
            logEncoding = self.logEncoding
        if not logEncoding:
            assert self.master is not None
            logEncoding = self.master.config.logEncoding
        log = plog.Log.new(self.master, name, type, logid, logEncoding)
        self.logs[name] = log
        self._connectPendingLogObservers()
        return log

    def _connectPendingLogObservers(self) -> None:
        for logname, observer in self._pendingLogObservers[:]:
            if logname in self.logs:
                observer.setLog(self.logs[logname])
                self._pendingLogObservers.remove((logname, observer))

    @defer.inlineCallbacks
    def addURL(self, name: str, url: str) -> InlineCallbacksType[None]:
        assert self.master is not None
        yield self.master.data.updates.addStepURL(self.stepid, str(name), str(url))
        return None

    @defer.inlineCallbacks
    def runCommand(self, command: remotecommand.RemoteCommand) -> InlineCallbacksType[int]:
        if self.stopped:
            return CANCELLED

        self.cmd = command
        command.worker = self.worker
        try:
            assert self.build
            assert self.build.builder.name
            res = yield command.run(self, self.remote, self.build.builder.name)
            if command.remote_failure_reason in ("timeout", "timeout_without_output"):
                self.timed_out = True
            elif command.remote_failure_reason in ("max_lines_failure",):
                self.max_lines_reached = True
        finally:
            self.cmd = None
        return res

    def hasStatistic(self, name: str) -> bool:
        return name in self.statistics

    def getStatistic(self, name: str, default: int | None = None) -> Any:
        return self.statistics.get(name, default)

    def getStatistics(self) -> dict[str, int]:
        return self.statistics.copy()

    def setStatistic(self, name: str, value: int) -> None:
        self.statistics[name] = value


class CommandMixin:
    getLog: Callable[[str], plog.Log]
    runCommand: Callable[[remotecommand.RemoteCommand], defer.Deferred[int]]

    @defer.inlineCallbacks
    def _runRemoteCommand(
        self,
        cmd: str | list[str],
        abandonOnFailure: bool,
        args: dict[str, Any],
        makeResult: Callable[[remotecommand.RemoteCommand], Any] | None = None,
    ) -> InlineCallbacksType[Any]:
        command = remotecommand.RemoteCommand(cmd, args)
        try:
            log = self.getLog('stdio')
        except Exception:
            log = yield self.addLog('stdio')  # type: ignore[attr-defined]
        command.useLog(log, False)
        yield self.runCommand(command)
        if abandonOnFailure and command.didFail():
            raise BuildStepFailed()
        if makeResult:
            return makeResult(command)
        else:
            return not command.didFail()

    def runRmdir(
        self, dir: str, log: str | None = None, abandonOnFailure: bool = True
    ) -> defer.Deferred[bool]:
        return self._runRemoteCommand('rmdir', abandonOnFailure, {'dir': dir, 'logEnviron': False})

    def pathExists(self, path: str) -> defer.Deferred[bool]:
        return self._runRemoteCommand('stat', False, {'file': path, 'logEnviron': False})

    def runMkdir(
        self, dir: str, log: str | None = None, abandonOnFailure: bool = True
    ) -> defer.Deferred[bool]:
        return self._runRemoteCommand('mkdir', abandonOnFailure, {'dir': dir, 'logEnviron': False})

    def runGlob(self, path: str) -> defer.Deferred[str]:
        return self._runRemoteCommand(
            'glob',
            True,
            {'path': path, 'logEnviron': False},
            makeResult=lambda cmd: cmd.updates['files'][0],
        )


class ShellMixin:
    command: list[str] | str | None = None
    env: dict[str, str] = {}
    want_stdout = True
    want_stderr = True
    usePTY: bool | None = None
    logfiles: dict[str, str] = {}
    lazylogfiles: bool = False
    timeout = 1200
    maxTime: float | None = None
    max_lines: int | None = None
    logEnviron = True
    interruptSignal = 'KILL'
    sigtermTime: int | None = None
    initialStdin: str | None = None
    decodeRC = {0: SUCCESS}

    getLog: Callable[[str], plog.Log]
    build: Build | None
    workdir: str | None
    workerVersionIsOlderThan: Callable[[str, str], bool]
    descriptionDone: str | list[str] | None
    results: int | None
    timed_out: bool
    max_lines_reached: bool

    _shell_mixin_arg_config = [
        ('command', None),
        ('workdir', check_param_str),
        ('env', None),
        ('want_stdout', check_param_bool),
        ('want_stderr', check_param_bool),
        ('usePTY', check_param_bool),
        ('logfiles', None),
        ('lazylogfiles', check_param_bool),
        ('timeout', check_param_number_none),
        ('maxTime', check_param_number_none),
        ('max_lines', check_param_number_none),
        ('logEnviron', check_param_bool),
        ('interruptSignal', check_param_str_none),
        ('sigtermTime', check_param_number_none),
        ('initialStdin', check_param_str_none),
        ('decodeRC', None),
    ]
    renderables: Sequence[str] = [arg for arg, _ in _shell_mixin_arg_config]

    def setupShellMixin(
        self, constructorArgs: dict[str, Any], prohibitArgs: list[str] | None = None
    ) -> dict[str, Any]:
        constructorArgs = constructorArgs.copy()

        if prohibitArgs is None:
            prohibitArgs = []

        def bad(arg: str) -> None:
            config.error(f"invalid {self.__class__.__name__} argument {arg}")

        for arg, check in self._shell_mixin_arg_config:
            if arg not in constructorArgs:
                continue
            if arg in prohibitArgs:
                bad(arg)
            else:
                value = constructorArgs[arg]
                if check is not None and not IRenderable.providedBy(value):
                    check(value, self.__class__, arg)

                setattr(self, arg, constructorArgs[arg])
            del constructorArgs[arg]
        for arg in list(constructorArgs):
            if arg not in BuildStep._params_names:
                bad(arg)
                del constructorArgs[arg]
        return constructorArgs

    @defer.inlineCallbacks
    def makeRemoteShellCommand(
        self,
        collectStdout: bool = False,
        collectStderr: bool = False,
        stdioLogName: str = 'stdio',
        **overrides: Any,
    ) -> InlineCallbacksType[remotecommand.RemoteShellCommand]:
        kwargs = {arg: getattr(self, arg) for arg, _ in self._shell_mixin_arg_config}
        kwargs.update(overrides)
        stdio = None
        if stdioLogName is not None:
            # Reuse an existing log if possible; otherwise, create one.
            try:
                stdio = yield self.getLog(stdioLogName)  # type: ignore[misc]
            except KeyError:
                stdio = yield self.addLog(stdioLogName)  # type: ignore[attr-defined]

        kwargs['command'] = flatten(kwargs['command'], (list, tuple))

        # store command away for display
        self.command = kwargs['command']

        # check for the usePTY flag
        if kwargs['usePTY'] is not None:
            if self.workerVersionIsOlderThan("shell", "2.7"):
                if stdio is not None:
                    yield stdio.addHeader("NOTE: worker does not allow master to override usePTY\n")
                del kwargs['usePTY']

        # check for the interruptSignal flag
        if kwargs["interruptSignal"] and self.workerVersionIsOlderThan("shell", "2.15"):
            if stdio is not None:
                yield stdio.addHeader(
                    "NOTE: worker does not allow master to specify interruptSignal\n"
                )
            del kwargs['interruptSignal']

        # lazylogfiles are handled below
        del kwargs['lazylogfiles']

        # merge the build's environment with that supplied here
        assert self.build is not None
        kwargs['env'] = {
            **(yield self.build.render(self.build.env)),
            **kwargs['env'],
        }
        kwargs['stdioLogName'] = stdioLogName

        # default the workdir appropriately
        if not kwargs.get('workdir') and not self.workdir:
            if callable(self.build.workdir):
                kwargs['workdir'] = self.build.workdir(self.build.sources)
            else:
                kwargs['workdir'] = self.build.workdir

        # the rest of the args go to RemoteShellCommand
        cmd = remotecommand.RemoteShellCommand(
            collectStdout=collectStdout, collectStderr=collectStderr, **kwargs
        )

        # set up logging
        if stdio is not None:
            cmd.useLog(stdio, False)
        for logname in self.logfiles:
            if self.lazylogfiles:
                # it's OK if this does, or does not, return a Deferred
                def callback(
                    cmd_arg: Any, local_logname: str = logname
                ) -> defer.Deferred[StreamLog]:
                    return self.addLog(local_logname)  # type: ignore[attr-defined]

                cmd.useLogDelayed(logname, callback, True)
            else:
                # add a LogFile
                newlog = yield self.addLog(logname)  # type: ignore[attr-defined]
                # and tell the RemoteCommand to feed it
                cmd.useLog(newlog, False)

        return cmd

    def getResultSummary(self) -> dict[str, str]:
        if self.descriptionDone is not None:
            return super().getResultSummary()  # type: ignore[misc]
        summary = util.command_to_string(self.command)
        if summary:
            if self.results != SUCCESS:
                summary += f' ({statusToString(self.results)})'
                if self.timed_out:
                    summary += " (timed out)"
                elif self.max_lines_reached:
                    summary += " (max lines)"

            if self.build is not None:
                summary = self.build.properties.cleanupTextFromSecrets(summary)
            return {'step': summary}
        return super().getResultSummary()  # type: ignore[misc]
